from __future__ import annotations

import asyncio
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Set, Tuple

import rich_click as click
from click.core import Context

if TYPE_CHECKING:
    from woke.config import WokeConfig


async def compile(
    config: WokeConfig,
    paths: Tuple[str],
    no_artifacts: bool,
    no_warnings: bool,
    force: bool,
    watch: bool,
    incremental: Optional[bool],
):
    from watchdog.observers import Observer

    from woke.compiler.compiler import (
        CompilationFileSystemEventHandler,
        SolidityCompiler,
    )
    from woke.compiler.solc_frontend.input_data_model import SolcOutputSelectionEnum

    from ..compiler.solc_frontend import SolcOutputErrorSeverityEnum
    from ..utils.file_utils import is_relative_to
    from .console import console

    compiler = SolidityCompiler(config)

    sol_files: Set[Path] = set()
    start = time.perf_counter()
    with console.status("[bold green]Searching for *.sol files...[/]"):
        if len(paths) == 0:
            for file in config.project_root_path.rglob("**/*.sol"):
                if (
                    not any(
                        is_relative_to(file, p)
                        for p in config.compiler.solc.ignore_paths
                    )
                    and file.is_file()
                ):
                    sol_files.add(file)
        else:
            for p in paths:
                path = Path(p)
                if path.is_file():
                    if not path.match("*.sol"):
                        raise ValueError(f"Argument `{p}` is not a Solidity file.")
                    sol_files.add(path)
                elif path.is_dir():
                    for file in path.rglob("**/*.sol"):
                        if (
                            not any(
                                is_relative_to(file, p)
                                for p in config.compiler.solc.ignore_paths
                            )
                            and file.is_file()
                        ):
                            sol_files.add(file)
                else:
                    raise ValueError(f"Argument `{p}` is not a file or directory.")
    end = time.perf_counter()
    console.log(
        f"[green]Found {len(sol_files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
    )

    if watch:
        fs_handler = CompilationFileSystemEventHandler(
            config,
            sol_files,
            asyncio.get_event_loop(),
            compiler,
            [SolcOutputSelectionEnum.ALL],
            write_artifacts=not no_artifacts,
            console=console,
            no_warnings=no_warnings,
        )

        observer = Observer()
        observer.schedule(
            fs_handler,
            str(config.project_root_path),
            recursive=True,
        )
        observer.start()
    else:
        fs_handler = None
        observer = None

    if not force:
        compiler.load(console=console)

    # TODO Allow choosing build artifacts subset in compile subcommand
    _, errors = await compiler.compile(
        sol_files,
        [SolcOutputSelectionEnum.ALL],
        write_artifacts=not no_artifacts,
        force_recompile=force,
        console=console,
        no_warnings=no_warnings,
        incremental=incremental,
    )

    if watch:
        assert fs_handler is not None
        assert observer is not None
        try:
            await fs_handler.run()
        except KeyboardInterrupt:
            pass
        finally:
            observer.stop()
            observer.join()
    else:
        errored = any(
            error.severity == SolcOutputErrorSeverityEnum.ERROR for error in errors
        )
        if errored:
            sys.exit(1)


@click.command(name="compile")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--no-artifacts", is_flag=True, default=False, help="Do not write build artifacts."
)
@click.option(
    "--no-warnings",
    is_flag=True,
    default=False,
    help="Do not print compilation warnings.",
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    default=False,
    help="Force recompile the project without previous build artifacts.",
)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    default=False,
    help="Watch for changes in the project and recompile on change.",
)
@click.option(
    "--incremental/--no-incremental",
    is_flag=True,
    required=False,
    default=None,
    help="Enforce incremental or non-incremental compilation.",
)
@click.pass_context
def run_compile(
    ctx: Context,
    paths: Tuple[str],
    no_artifacts: bool,
    no_warnings: bool,
    force: bool,
    watch: bool,
    incremental: Optional[bool],
) -> None:
    """Compile the project."""
    from woke.config import WokeConfig

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    asyncio.run(
        compile(config, paths, no_artifacts, no_warnings, force, watch, incremental)
    )



try:
    from importlib import metadata
except ImportError:
    # Python < 3.8
    import importlib_metadata as metadata  # pyright: ignore reportMissingImports


def get_package_version(package_name: str) -> str:
    """Get the version of the given package."""
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        raise RuntimeError(f"Package {package_name} not found.")



import logging
from typing import List, Union

from woke.ir import (
    BinaryOperation,
    DeclarationAbc,
    ExternalReference,
    Identifier,
    IdentifierPath,
    IdentifierPathPart,
    IrAbc,
    MemberAccess,
    UnaryOperation,
    UserDefinedTypeName,
    YulIdentifier,
)
from woke.ir.enums import GlobalSymbolsEnum
from woke.lsp.common_structures import (
    DocumentUri,
    Location,
    PartialResultParams,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path

logger = logging.getLogger(__name__)


class ReferenceOptions(WorkDoneProgressOptions):
    pass


class ReferenceRegistrationOptions(TextDocumentRegistrationOptions, ReferenceOptions):
    pass


class ReferenceContext(LspModel):
    include_declaration: bool


class ReferenceParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    context: ReferenceContext


def _generate_reference_location(
    reference: Union[
        Identifier,
        IdentifierPathPart,
        MemberAccess,
        ExternalReference,
        UnaryOperation,
        BinaryOperation,
        DeclarationAbc,
    ],
    context: LspContext,
) -> Location:
    path = reference.file
    if isinstance(reference, MemberAccess):
        location = reference.member_byte_location
    elif isinstance(reference, ExternalReference):
        location = reference.identifier_byte_location
    elif isinstance(reference, DeclarationAbc):
        location = reference.name_location
    else:
        location = reference.byte_location

    return Location(
        uri=DocumentUri(path_to_uri(path)),
        range=context.compiler.get_range_from_byte_offsets(path, location),
    )


async def references(
    context: LspContext, params: ReferenceParams
) -> Union[List[Location], None]:
    logger.debug(
        f"References for file {params.text_document.uri} at position {params.position} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, DeclarationAbc):
        name_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.name_location
        )
        if not position_within_range(params.position, name_location_range):
            return None

    if isinstance(node, (Identifier, MemberAccess)):
        referenced_declaration = node.referenced_declaration
        if referenced_declaration is None:
            return None
        if isinstance(referenced_declaration, GlobalSymbolsEnum):
            global_refs = (
                context.compiler.ir_reference_resolver.get_global_symbol_references(
                    referenced_declaration
                )
            )
            if len(global_refs) > 0:
                return [
                    _generate_reference_location(ref, context) for ref in global_refs
                ]
            return None
        else:
            node = referenced_declaration
    elif isinstance(node, (IdentifierPath, UserDefinedTypeName)):
        part = node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(node, YulIdentifier):
        external_reference = node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration
    elif (
        isinstance(node, (UnaryOperation, BinaryOperation))
        and node.function is not None
    ):
        node = node.function

    if not isinstance(node, DeclarationAbc):
        return None

    refs = []
    for reference in node.get_all_references(
        context.config.lsp.find_references.include_declarations
    ):
        refs.append(_generate_reference_location(reference, context))

    if len(refs) == 0:
        return None
    return refs



import sys
from typing import List, Tuple

import click.shell_completion as shell_completion
import rich_click as click


class FileAndPassParamType(click.ParamType):
    name = "file_and_pass"

    def convert(self, value, param, ctx):
        return value

    def shell_complete(
        self, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List:
        return [shell_completion.CompletionItem(incomplete, type="file")]


@click.command(
    name="test",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option(
    "--debug", "-d", is_flag=True, default=False, help="Attach debugger on exception."
)
@click.option(
    "--s/--no-s",
    "-s",
    is_flag=True,
    default=True,
    help="Show stdout and stderr of test.",
)
@click.option(
    "--coverage/--no-coverage",
    is_flag=True,
    default=False,
    help="Create coverage report.",
)
@click.argument("pytest_args", nargs=-1, type=FileAndPassParamType())
@click.pass_context
def run_test(
    context: click.Context,
    debug: bool,
    s: bool,
    coverage: bool,
    pytest_args: Tuple[str, ...],
) -> None:
    """Execute Woke tests using pytest."""
    import pytest

    from woke.config import WokeConfig
    from woke.development.globals import chain_interfaces_manager, set_config
    from woke.testing.pytest_plugin import PytestWokePlugin

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    set_config(config)

    if debug:
        from woke.development.globals import attach_debugger, set_exception_handler

        set_exception_handler(attach_debugger)

    if coverage:
        from woke.development.globals import set_coverage_handler
        from woke.testing.coverage import CoverageHandler

        set_coverage_handler(CoverageHandler(config))

    args = list(pytest_args)
    if s:
        args.append("-s")

    # disable pytest-brownie
    args.append("-p")
    args.append("no:pytest-brownie")

    # disable ape_test
    args.append("-p")
    args.append("no:ape_test")

    # disable pytest_ethereum
    args.append("-p")
    args.append("no:pytest_ethereum")

    ret = pytest.main(args, plugins=[PytestWokePlugin(config)])

    if coverage:
        from woke.development.globals import get_coverage_handler
        from woke.testing.coverage import export_merged_ide_coverage, write_coverage

        coverage_handler = get_coverage_handler()
        assert coverage_handler is not None

        c = export_merged_ide_coverage([coverage_handler.get_contract_ide_coverage()])
        write_coverage(c, config.project_root_path / "woke-coverage.cov")

    chain_interfaces_manager.close_all()

    sys.exit(ret)



import asyncio
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Set, Type, Union

import rich_click as click

if TYPE_CHECKING:
    from woke.printers import Printer


class PrintCli(click.RichGroup):  # pyright: ignore reportPrivateImportUsage
    _plugins_loaded = False
    _plugin_commands: Dict[str, click.Command] = {}

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[
            Union[Dict[str, click.Command], Sequence[click.Command]]
        ] = None,
        **attrs: Any,
    ):
        super().__init__(name=name, commands=commands, **attrs)

        for command in self.commands.values():
            self._inject_params(command)

    @staticmethod
    def _inject_params(command: click.Command) -> None:
        command.params.append(
            click.Argument(
                ["paths"],
                nargs=-1,
                type=click.Path(exists=True),
            )
        )

    def _load_plugins(self) -> None:
        if sys.version_info < (3, 10):
            from importlib_metadata import entry_points
        else:
            from importlib.metadata import entry_points

        printer_entry_points = entry_points().select(group="woke.plugins.printers")
        for entry_point in printer_entry_points:
            entry_point.load()

        if (
            Path.cwd().joinpath("printers").is_dir()
            and Path.cwd().joinpath("printers/__init__.py").is_file()
        ):
            from importlib.util import module_from_spec, spec_from_file_location

            sys.path.insert(0, str(Path.cwd()))
            spec = spec_from_file_location("printers", "printers/__init__.py")
            if spec is not None and spec.loader is not None:
                module = module_from_spec(spec)
                spec.loader.exec_module(module)

        self._plugins_loaded = True

    def add_command(self, cmd: click.Command, name: Optional[str] = None) -> None:
        self._inject_params(cmd)
        super().add_command(cmd, name)

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        if not self._plugins_loaded:
            self._load_plugins()
        return self.commands.get(cmd_name)

    def list_commands(self, ctx: click.Context) -> List[str]:
        if not self._plugins_loaded:
            self._load_plugins()
        return sorted(self.commands)

    def invoke(self, ctx: click.Context):
        ctx.obj["subcommand_args"] = ctx.args
        ctx.obj["subcommand_protected_args"] = ctx.protected_args
        super().invoke(ctx)


@click.group(
    name="print", cls=PrintCli, context_settings={"auto_envvar_prefix": "WOKE_PRINTER"}
)
@click.option(
    "--no-artifacts", is_flag=True, default=False, help="Do not write build artifacts."
)
@click.pass_context
def run_print(ctx: click.Context, no_artifacts: bool) -> None:
    """Run a printer."""

    if "--help" in ctx.obj["subcommand_args"]:
        return

    from ..compiler import SolcOutputSelectionEnum, SolidityCompiler
    from ..compiler.build_data_model import ProjectBuild
    from ..compiler.solc_frontend import SolcOutputError, SolcOutputErrorSeverityEnum
    from ..config import WokeConfig
    from ..utils import get_class_that_defined_method
    from ..utils.file_utils import is_relative_to
    from .console import console

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    sol_files: Set[Path] = set()
    start = time.perf_counter()
    with console.status("[bold green]Searching for *.sol files...[/]"):
        for file in config.project_root_path.rglob("**/*.sol"):
            if (
                not any(
                    is_relative_to(file, p) for p in config.compiler.solc.ignore_paths
                )
                and file.is_file()
            ):
                sol_files.add(file)
    end = time.perf_counter()
    console.log(
        f"[green]Found {len(sol_files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
    )

    compiler = SolidityCompiler(config)
    compiler.load(console=console)

    build: ProjectBuild
    errors: Set[SolcOutputError]
    build, errors = asyncio.run(
        compiler.compile(
            sol_files,
            [SolcOutputSelectionEnum.ALL],
            write_artifacts=not no_artifacts,
            console=console,
            no_warnings=True,
        )
    )

    errored = any(
        error.severity == SolcOutputErrorSeverityEnum.ERROR for error in errors
    )
    if errored:
        sys.exit(1)

    assert compiler.latest_build_info is not None
    assert compiler.latest_graph is not None

    assert isinstance(ctx.command, PrintCli)
    assert ctx.invoked_subcommand is not None
    command = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    assert command is not None
    assert command.name is not None

    if hasattr(config.printers, command.name):
        default_map = getattr(config.printers, command.name)
    else:
        default_map = None

    cls: Type[Printer] = get_class_that_defined_method(
        command.callback
    )  # pyright: ignore reportGeneralTypeIssues
    if cls is not None:

        def _callback(*args, **kwargs):
            instance.paths = [Path(p).resolve() for p in kwargs.pop("paths", [])]

            original_callback(
                instance, *args, **kwargs
            )  # pyright: ignore reportOptionalCall

        instance = cls()
        instance.build = build
        instance.build_info = compiler.latest_build_info
        instance.config = config
        instance.console = console
        instance.imports_graph = (  # pyright: ignore reportGeneralTypeIssues
            compiler.latest_graph.copy()
        )

        original_callback = command.callback
        command.callback = _callback

        sub_ctx = command.make_context(
            command.name,
            [*ctx.obj["subcommand_protected_args"][1:], *ctx.obj["subcommand_args"]],
            parent=ctx,
            default_map=default_map,
        )
        with sub_ctx:
            sub_ctx.command.invoke(sub_ctx)

        instance._run()
    else:

        def _callback(*args, **kwargs):
            click.get_current_context().obj["paths"] = [
                Path(p).resolve() for p in kwargs.pop("paths", [])
            ]

            original_callback(*args, **kwargs)  # pyright: ignore reportOptionalCall

        args = [*ctx.obj["subcommand_protected_args"][1:], *ctx.obj["subcommand_args"]]
        ctx.obj = {
            "build": build,
            "build_info": compiler.latest_build_info,
            "config": config,
            "console": console,
            "imports_graph": compiler.latest_graph.copy(),
        }

        original_callback = command.callback
        command.callback = _callback

        sub_ctx = command.make_context(
            command.name, args, parent=ctx, default_map=default_map
        )
        with sub_ctx:
            sub_ctx.command.invoke(sub_ctx)

    # avoid double execution of a subcommand
    sys.exit(0)



from __future__ import annotations

import functools
import importlib
import inspect
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field, fields
from enum import IntEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.error import HTTPError

from typing_extensions import get_args, get_origin, get_type_hints

if TYPE_CHECKING:
    from .blocks import Block

from .call_trace import CallTrace
from .chain_interfaces import (
    AnvilChainInterface,
    GanacheChainInterface,
    GethLikeChainInterfaceAbc,
    HardhatChainInterface,
    TxParams,
)
from .core import (
    Account,
    Address,
    Chain,
    Wei,
    get_contract_from_fqn,
    get_fqn_from_address,
    get_fqn_from_creation_code,
)
from .internal import UnknownEvent, read_from_memory
from .json_rpc import JsonRpcError

T = TypeVar("T")


class TransactionStatusEnum(IntEnum):
    PENDING = -1
    SUCCESS = 1
    FAILURE = 0


class TransactionTypeEnum(IntEnum):
    LEGACY = 0
    EIP2930 = 1
    EIP1559 = 2


def _fetch_tx_data(f):
    @functools.wraps(f)
    def wrapper(self: TransactionAbc):
        if self._tx_data is None:
            self._tx_data = self._chain.chain_interface.get_transaction(self.tx_hash)
        return f(self)

    return wrapper


def _fetch_tx_receipt(f):
    @functools.wraps(f)
    def wrapper(self: TransactionAbc):
        if self._tx_receipt is None:
            self.wait()
        assert self._tx_receipt is not None
        return f(self)

    return wrapper


class ChainTransactions:
    _chain: Chain
    _transactions: Dict[str, TransactionAbc]

    def __init__(self, chain: Chain):
        self._chain = chain
        self._transactions = {}

    def __getitem__(self, key: str) -> TransactionAbc:
        if not key.startswith("0x"):
            key = "0x" + key
        key = key.lower()
        if key in self._transactions:
            return self._transactions[key]

        tx_data = self._chain.chain_interface.get_transaction(key)
        type = int(tx_data["type"], 16)

        tx_params: TxParams = {
            "nonce": int(tx_data["nonce"], 16),
            "from": tx_data["from"],
            "gas": int(tx_data["gas"], 16),
            "value": int(tx_data["value"], 16),
            "data": bytes.fromhex(tx_data["input"][2:]),
        }

        if "to" in tx_data and tx_data["to"] is not None:
            tx_params["to"] = tx_data["to"]
            try:
                fqn = get_fqn_from_address(
                    Address(tx_params["to"]),
                    int(tx_data["blockNumber"], 16) - 1
                    if "blockNumber" in tx_data
                    else "latest",
                    self._chain,
                )
                module_name, attrs = get_contract_from_fqn(
                    fqn  # pyright: ignore reportGeneralTypeIssues
                )
                obj = getattr(importlib.import_module(module_name), attrs[0])
                for attr in attrs[1:]:
                    obj = getattr(obj, attr)
                selector = tx_params["data"][:4]
                abi = obj._abi[selector]
                method = next(
                    getattr(obj, m)
                    for m in dir(obj)
                    if hasattr(getattr(obj, m), "selector")
                    and getattr(obj, m).selector == selector
                )
                return_types = get_args(get_type_hints(method)["return"])
                return_type = next(
                    get_args(t)[0]
                    for t in return_types
                    if get_origin(t) is TransactionAbc
                )
            except (KeyError, StopIteration):
                abi = None
                return_type = bytearray
        else:
            try:
                fqn = get_fqn_from_creation_code(tx_params["data"])[0]
                module_name, attrs = get_contract_from_fqn(fqn)
                obj = getattr(importlib.import_module(module_name), attrs[0])
                for attr in attrs[1:]:
                    obj = getattr(obj, attr)
                if "constructor" in obj._abi:
                    abi = obj._abi["constructor"]
                else:
                    abi = None
                return_type = obj
            except (KeyError, ValueError):
                abi = None
                return_type = Account

        if type == 0:
            tx_params["gasPrice"] = int(tx_data["gasPrice"], 16)
            tx = LegacyTransaction(key, tx_params, abi, return_type, self._chain)
        elif type == 1:
            tx_params["gasPrice"] = int(tx_data["gasPrice"], 16)
            tx_params["accessList"] = tx_data["accessList"]
            tx_params["chainId"] = 1
            tx = Eip2930Transaction(key, tx_params, abi, return_type, self._chain)
        elif type == 2:
            tx_params["maxFeePerGas"] = int(tx_data["maxFeePerGas"], 16)
            tx_params["maxPriorityFeePerGas"] = int(tx_data["maxPriorityFeePerGas"], 16)
            tx_params["accessList"] = tx_data["accessList"]
            tx_params["chainId"] = 2
            tx = Eip1559Transaction(key, tx_params, abi, return_type, self._chain)
        else:
            raise ValueError(f"Unknown transaction type {type}")

        self._transactions[key] = tx
        return tx


class TransactionAbc(ABC, Generic[T]):
    _tx_hash: str
    _tx_params: TxParams
    _chain: Chain
    _abi: Optional[Dict]
    _return_type: Type
    _tx_data: Optional[Dict[str, Any]]
    _tx_receipt: Optional[Dict[str, Any]]
    _trace_transaction: Optional[List[Dict[str, Any]]]
    _debug_trace_transaction = Optional[Dict[str, Any]]
    _error: Optional[TransactionRevertedError]
    _raw_error: Optional[UnknownTransactionRevertedError]
    _events: Optional[List]

    def __init__(
        self,
        tx_hash: str,
        tx_params: TxParams,
        abi: Optional[Dict],
        return_type: Type,
        chain: Chain,
    ):
        self._tx_hash = tx_hash
        self._tx_params = tx_params
        self._abi = abi
        self._return_type = return_type
        self._chain = chain

        self._tx_data = None
        self._tx_receipt = None
        self._trace_transaction = None
        self._debug_trace_transaction = None
        self._error = None
        self._raw_error = None
        self._events = None

    @property
    def tx_hash(self) -> str:
        return self._tx_hash

    @property
    def chain(self) -> Chain:
        return self._chain

    @property
    @_fetch_tx_receipt
    def block(self) -> Block:
        return self._chain.blocks[self.block_number]

    @property
    @_fetch_tx_receipt
    def block_number(self) -> int:
        return int(
            self._tx_receipt["blockNumber"],  # pyright: ignore reportOptionalSubscript
            16,
        )

    @property
    def data(self) -> bytes:
        return self._tx_params["data"] if "data" in self._tx_params else b""

    @property
    def from_(self) -> Account:
        return Account(
            self._tx_params["from"],  # pyright: ignore reportTypedDictNotRequiredAccess
            self._chain,
        )

    @property
    def to(self) -> Optional[Account]:
        if "to" in self._tx_params:
            return Account(self._tx_params["to"], self._chain)
        return None

    @property
    def gas_limit(self) -> int:
        return self._tx_params[
            "gas"
        ]  # pyright: ignore reportTypedDictNotRequiredAccess

    @property
    def nonce(self) -> int:
        return self._tx_params[
            "nonce"
        ]  # pyright: ignore reportTypedDictNotRequiredAccess

    @property
    @_fetch_tx_receipt
    def tx_index(self) -> int:
        return int(
            self._tx_receipt[  # pyright: ignore reportOptionalSubscript
                "transactionIndex"
            ],
            16,
        )

    @property
    def value(self) -> Wei:
        return Wei(
            self._tx_params["value"]  # pyright: ignore reportTypedDictNotRequiredAccess
        )

    @property
    @_fetch_tx_data
    def r(self) -> int:
        return int(self._tx_data["r"], 16)  # pyright: ignore reportOptionalSubscript

    @property
    @_fetch_tx_data
    def s(self) -> int:
        return int(self._tx_data["s"], 16)  # pyright: ignore reportOptionalSubscript

    @property
    @_fetch_tx_receipt
    def gas_used(self) -> int:
        return int(
            self._tx_receipt["gasUsed"], 16  # pyright: ignore reportOptionalSubscript
        )

    @property
    @_fetch_tx_receipt
    def cumulative_gas_used(self) -> int:
        return int(
            self._tx_receipt[  # pyright: ignore reportOptionalSubscript
                "cumulativeGasUsed"
            ],
            16,
        )

    @property
    @_fetch_tx_receipt
    def effective_gas_price(self) -> Wei:
        return Wei(
            int(
                self._tx_receipt[  # pyright: ignore reportOptionalSubscript
                    "effectiveGasPrice"
                ],
                16,
            )
        )

    @property
    def status(self) -> TransactionStatusEnum:
        if self._tx_receipt is None:
            receipt = self._chain.chain_interface.get_transaction_receipt(self._tx_hash)
            if receipt is None:
                return TransactionStatusEnum.PENDING
            else:
                self._tx_receipt = receipt

        if int(self._tx_receipt["status"], 16) == 0:
            return TransactionStatusEnum.FAILURE
        else:
            return TransactionStatusEnum.SUCCESS

    def wait(self, confirmations: Optional[int] = None) -> None:
        self._chain._wait_for_transaction(self, confirmations)

    def _fetch_trace_transaction(self) -> None:
        if self._trace_transaction is None:
            chain_interface = self._chain.chain_interface
            self._trace_transaction = chain_interface.trace_transaction(self._tx_hash)

    def _fetch_debug_trace_transaction(self) -> None:
        if self._debug_trace_transaction is None:
            self._debug_trace_transaction = (
                self._chain.chain_interface.debug_trace_transaction(
                    self._tx_hash,
                    {"enableMemory": True},
                )
            )

    @property
    @_fetch_tx_receipt
    def console_logs(self) -> list:
        chain_interface = self._chain.chain_interface

        if isinstance(chain_interface, AnvilChainInterface):
            self._fetch_trace_transaction()
            assert self._trace_transaction is not None
            return self._chain._process_console_logs(self._trace_transaction)
        elif isinstance(
            chain_interface, (GanacheChainInterface, HardhatChainInterface)
        ):
            self._fetch_debug_trace_transaction()
            assert self._debug_trace_transaction is not None
            return self._chain._process_console_logs_from_debug_trace(
                self._debug_trace_transaction  # pyright: ignore reportGeneralTypeIssues
            )
        elif isinstance(chain_interface, GethLikeChainInterfaceAbc):
            try:
                self._fetch_trace_transaction()
                assert self._trace_transaction is not None
                return self._chain._process_console_logs(self._trace_transaction)
            except (JsonRpcError, HTTPError):
                # TODO make assertions about error.code?
                try:
                    self._fetch_debug_trace_transaction()
                    assert self._debug_trace_transaction is not None
                    return self._chain._process_console_logs_from_debug_trace(
                        self._debug_trace_transaction  # pyright: ignore reportGeneralTypeIssues
                    )
                except (JsonRpcError, HTTPError):
                    # TODO make assertions about error.code?
                    raise RuntimeError(
                        f"Could not get console logs for transaction {self.tx_hash} as trace_transaction and debug_trace_transaction are both unavailable"
                    )
        else:
            raise NotImplementedError

    @property
    @_fetch_tx_receipt
    def events(self) -> list:
        if self._events is not None:
            return self._events

        assert self._tx_receipt is not None

        if len(self._tx_receipt["logs"]) == 0:
            self._events = []
            return self._events

        self._events = self._chain._process_events(self, self._tx_receipt["logs"])
        return self._events

    @property
    @_fetch_tx_receipt
    def raw_events(self) -> List[UnknownEvent]:
        assert self._tx_receipt is not None

        ret = []
        for log in self._tx_receipt["logs"]:
            topics = [
                bytes.fromhex(t[2:]) if t.startswith("0x") else bytes.fromhex(t)
                for t in log["topics"]
            ]
            data = (
                bytes.fromhex(log["data"][2:])
                if log["data"].startswith("0x")
                else bytes.fromhex(log["data"])
            )
            ret.append(UnknownEvent(topics, data))
        return ret

    @property
    @_fetch_tx_receipt
    def error(self) -> Optional[TransactionRevertedError]:
        if self.status == TransactionStatusEnum.SUCCESS:
            return None

        if self._error is not None:
            return self._error

        raw_error = self.raw_error
        assert raw_error is not None

        try:
            self._chain._process_revert_data(self, raw_error.data)
        except TransactionRevertedError as e:
            self._error = e
            return e

    @property
    @_fetch_tx_receipt
    def raw_error(self) -> Optional[UnknownTransactionRevertedError]:
        if self.status == TransactionStatusEnum.SUCCESS:
            return None

        if self._raw_error is not None:
            return self._raw_error

        chain_interface = self._chain.chain_interface

        assert self._tx_receipt is not None
        # due to a bug, Anvil does not return revert data for failed contract creations
        # "contractAddress" key is always present, so the first if branch is never taken
        # keep this code for when the bug is fixed
        if (
            isinstance(chain_interface, AnvilChainInterface)
            and "contractAddress" not in self._tx_receipt
        ):
            self._fetch_trace_transaction()
            assert self._trace_transaction is not None
            revert_data = bytes.fromhex(
                self._trace_transaction[0]["result"]["output"][2:]
            )
        elif isinstance(chain_interface, (AnvilChainInterface, GanacheChainInterface)):
            self._fetch_debug_trace_transaction()
            assert self._debug_trace_transaction is not None

            if len(self._debug_trace_transaction["structLogs"]) == 0 or self._debug_trace_transaction["structLogs"][-1]["op"] != "REVERT":  # type: ignore
                revert_data = b""
            else:
                trace: Any = self._debug_trace_transaction["structLogs"][-1]  # type: ignore
                offset = int(trace["stack"][-1], 16)
                length = int(trace["stack"][-2], 16)

                revert_data = bytes(read_from_memory(offset, length, trace["memory"]))
        elif isinstance(chain_interface, HardhatChainInterface):
            self._fetch_debug_trace_transaction()
            assert self._debug_trace_transaction is not None
            revert_data = bytes.fromhex(self._debug_trace_transaction["returnValue"])  # type: ignore
        elif isinstance(chain_interface, GethLikeChainInterfaceAbc):
            try:
                self._fetch_trace_transaction()
                assert self._trace_transaction is not None
                revert_data = bytes.fromhex(
                    self._trace_transaction[0]["result"]["output"][2:]
                )
            except (JsonRpcError, HTTPError):
                # TODO make assertions about error.code?
                try:
                    self._fetch_debug_trace_transaction()
                    assert self._debug_trace_transaction is not None
                    revert_data = bytes.fromhex(self._debug_trace_transaction["returnValue"])  # type: ignore
                except (JsonRpcError, HTTPError):
                    # TODO make assertions about error.code?
                    raise RuntimeError(
                        f"Could not get revert reason data for transaction {self.tx_hash} as trace_transaction and debug_trace_transaction are both unavailable"
                    )
        else:
            raise NotImplementedError

        self._raw_error = UnknownTransactionRevertedError(revert_data)
        self._raw_error.tx = self
        return self._raw_error

    @property
    @_fetch_tx_receipt
    def return_value(self) -> T:
        raw_value = self.raw_return_value

        if self._return_type is type(None):
            return None  # pyright: ignore reportGeneralTypeIssues

        if isinstance(raw_value, Account):
            return self._return_type(raw_value.address, self._chain)
        elif isinstance(raw_value, bytearray):
            if self._abi is None:
                return self._return_type(raw_value)

            return self._chain._process_return_data(
                self, bytes(raw_value), self._abi, self._return_type
            )
        else:
            raise TypeError(
                f"Unexpected return type from transaction {self.tx_hash}: {type(raw_value)}"
            )

    @property
    @_fetch_tx_receipt
    def raw_return_value(self) -> Union[Account, bytearray]:
        if self.status != TransactionStatusEnum.SUCCESS:
            e = self.error
            assert e is not None
            raise e

        assert self._tx_receipt is not None
        if (
            "contractAddress" in self._tx_receipt
            and self._tx_receipt["contractAddress"] is not None
        ):
            return Account(self._tx_receipt["contractAddress"], self._chain)

        chain_interface = self._chain.chain_interface
        if isinstance(chain_interface, AnvilChainInterface):
            self._fetch_trace_transaction()
            assert self._trace_transaction is not None
            output = bytes.fromhex(self._trace_transaction[0]["result"]["output"][2:])
        elif isinstance(chain_interface, GanacheChainInterface):
            self._fetch_debug_trace_transaction()
            assert self._debug_trace_transaction is not None

            if len(self._debug_trace_transaction["structLogs"]) == 0 or self._debug_trace_transaction["structLogs"][-1]["op"] != "RETURN":  # type: ignore
                output = b""
            else:
                trace: Any = self._debug_trace_transaction["structLogs"][-1]  # type: ignore
                offset = int(trace["stack"][-1], 16)
                length = int(trace["stack"][-2], 16)

                output = read_from_memory(offset, length, trace["memory"])
        elif isinstance(chain_interface, HardhatChainInterface):
            self._fetch_debug_trace_transaction()
            assert self._debug_trace_transaction is not None
            output = bytes.fromhex(self._debug_trace_transaction["returnValue"])  # type: ignore
        elif isinstance(chain_interface, GethLikeChainInterfaceAbc):
            try:
                self._fetch_trace_transaction()
                assert self._trace_transaction is not None
                output = bytes.fromhex(
                    self._trace_transaction[0]["result"]["output"][2:]
                )
            except (JsonRpcError, HTTPError):
                # TODO make assertions about error.code?
                try:
                    self._fetch_debug_trace_transaction()
                    assert self._debug_trace_transaction is not None
                    output = bytes.fromhex(self._debug_trace_transaction["returnValue"])  # type: ignore
                except (JsonRpcError, HTTPError):
                    # TODO make assertions about error.code?
                    raise RuntimeError(
                        f"Could not get return value for transaction {self.tx_hash} as trace_transaction and debug_trace_transaction are both unavailable"
                    )
        else:
            raise NotImplementedError

        return bytearray(output)

    @property
    @_fetch_tx_receipt
    def call_trace(self) -> CallTrace:
        if self._debug_trace_transaction is None:
            self._fetch_debug_trace_transaction()
        assert self._debug_trace_transaction is not None

        return CallTrace.from_debug_trace(
            self,
            self._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
            self._tx_params,
        )

    @property
    @abstractmethod
    def type(self) -> TransactionTypeEnum:
        ...


class LegacyTransaction(TransactionAbc[T]):
    @property
    @_fetch_tx_data
    def v(self) -> int:
        return int(self._tx_data["v"], 16)  # pyright: ignore reportOptionalSubscript

    @property
    def gas_price(self) -> Wei:
        assert "gasPrice" in self._tx_params
        return Wei(self._tx_params["gasPrice"])

    @property
    def type(self) -> TransactionTypeEnum:
        return TransactionTypeEnum.LEGACY


class Eip2930Transaction(TransactionAbc[T]):
    @property
    def chain_id(self) -> int:
        assert "chainId" in self._tx_params
        return self._tx_params["chainId"]

    @property
    def access_list(self) -> Dict[Account, List[int]]:
        assert "accessList" in self._tx_params
        ret = {}
        for entry in self._tx_params["accessList"]:
            account = Account(entry[0])
            if account not in ret:
                ret[account] = []
            ret[account].append(entry[1])
        return ret

    @property
    def gas_price(self) -> Wei:
        assert "gasPrice" in self._tx_params
        return Wei(self._tx_params["gasPrice"])

    @property
    @_fetch_tx_data
    def y_parity(self) -> bool:
        return bool(
            int(self._tx_data["v"], 16) & 1  # pyright: ignore reportOptionalSubscript
        )

    @property
    def type(self) -> TransactionTypeEnum:
        assert "type" in self._tx_params and self._tx_params["type"] == 1
        return TransactionTypeEnum.EIP2930


class Eip1559Transaction(TransactionAbc[T]):
    @property
    def chain_id(self) -> int:
        assert "chainId" in self._tx_params
        return self._tx_params["chainId"]

    @property
    def max_fee_per_gas(self) -> Wei:
        if "maxFeePerGas" not in self._tx_params:
            if self._tx_data is None:
                self._tx_data = self._chain.chain_interface.get_transaction(
                    self.tx_hash
                )
            return Wei(int(self._tx_data["maxFeePerGas"], 16))
        return Wei(self._tx_params["maxFeePerGas"])

    @property
    def max_priority_fee_per_gas(self) -> Wei:
        if "maxPriorityFeePerGas" not in self._tx_params:
            if self._tx_data is None:
                self._tx_data = self._chain.chain_interface.get_transaction(
                    self.tx_hash
                )
            return Wei(int(self._tx_data["maxPriorityFeePerGas"], 16))
        return Wei(self._tx_params["maxPriorityFeePerGas"])

    @property
    def access_list(self) -> Dict[Account, List[int]]:
        assert "accessList" in self._tx_params
        ret = {}
        for entry in self._tx_params["accessList"]:
            account = Account(entry[0])
            if account not in ret:
                ret[account] = []
            ret[account].append(entry[1])
        return ret

    @property
    @_fetch_tx_data
    def y_parity(self) -> bool:
        return bool(
            int(self._tx_data["v"], 16) & 1  # pyright: ignore reportOptionalSubscript
        )

    @property
    def type(self) -> TransactionTypeEnum:
        assert "type" in self._tx_params and self._tx_params["type"] == 2
        return TransactionTypeEnum.EIP1559


@dataclass
class TransactionRevertedError(Exception):
    tx: Optional[TransactionAbc] = field(
        init=False, compare=False, default=None, repr=False
    )

    def __str__(self):
        s = ", ".join(
            [f"{f.name}={getattr(self, f.name)!r}" for f in fields(self) if f.init]
        )
        return f"{self.__class__.__qualname__}({s})"


@dataclass
class UnknownTransactionRevertedError(TransactionRevertedError):
    data: bytes


@dataclass
class Error(TransactionRevertedError):
    _abi = {
        "name": "Error",
        "type": "error",
        "inputs": [{"internalType": "string", "name": "message", "type": "string"}],
    }
    message: str


class PanicCodeEnum(IntEnum):
    GENERIC = 0
    "Generic compiler panic"
    ASSERT_FAIL = 1
    "Assert evaluated to false"
    UNDERFLOW_OVERFLOW = 0x11
    "Integer underflow or overflow"
    DIVISION_MODULO_BY_ZERO = 0x12
    "Division or modulo by zero"
    INVALID_CONVERSION_TO_ENUM = 0x21
    "Too big or negative integer for conversion to enum"
    ACCESS_TO_INCORRECTLY_ENCODED_STORAGE_BYTE_ARRAY = 0x22
    "Access to incorrectly encoded storage byte array"
    POP_EMPTY_ARRAY = 0x31
    ".pop() on empty array"
    INDEX_ACCESS_OUT_OF_BOUNDS = 0x32
    "Out-of-bounds or negative index access to fixed-length array"
    TOO_MUCH_MEMORY_ALLOCATED = 0x41
    "Too much memory allocated"
    INVALID_INTERNAL_FUNCTION_CALL = 0x51
    "Called invalid internal function"


@dataclass
class Panic(TransactionRevertedError):
    _abi = {
        "name": "Panic",
        "type": "error",
        "inputs": [{"internalType": "uint256", "name": "code", "type": "uint256"}],
    }
    code: "PanicCodeEnum"


class ExceptionWrapper:
    value: Optional[Exception] = None


@contextmanager
def must_revert(
    exceptions: Union[
        str,
        int,
        Exception,
        Type[Exception],
        Tuple[Union[str, int, Exception, Type[Exception]], ...],
    ] = TransactionRevertedError,
) -> Iterator[ExceptionWrapper]:
    if isinstance(exceptions, str):
        exceptions = Error(exceptions)
    elif isinstance(exceptions, int):
        exceptions = Panic(PanicCodeEnum(exceptions))

    if isinstance(exceptions, (tuple, list)):
        tmp: List[Union[str, int, Exception, Type[Exception]]] = []
        for ex in exceptions:
            if isinstance(ex, str):
                tmp.append(Error(ex))
            elif isinstance(ex, int):
                tmp.append(Panic(PanicCodeEnum(ex)))
            else:
                tmp.append(ex)
        exceptions = tuple(tmp)
        types = tuple(type(x) if not inspect.isclass(x) else x for x in exceptions)
    else:
        types = type(exceptions) if not inspect.isclass(exceptions) else exceptions

    wrapper = ExceptionWrapper()

    try:
        yield wrapper
        raise AssertionError(f"Expected revert of type {exceptions}")
    except types as e:  # pyright: ignore reportGeneralTypeIssues
        wrapper.value = e

        if isinstance(exceptions, (tuple, list)):
            if any(
                (inspect.isclass(ex) and issubclass(type(e), ex)) or e == ex
                for ex in exceptions
            ):
                return
            raise AssertionError(f"Expected revert of type {exceptions} but got {e}")
        else:
            if not inspect.isclass(exceptions):
                assert e == exceptions, f"Expected {e} but got {exceptions}"


@contextmanager
def may_revert(
    exceptions: Union[
        str,
        int,
        Exception,
        Type[Exception],
        Tuple[Union[str, int, Exception, Type[Exception]], ...],
    ] = TransactionRevertedError,
) -> Iterator[ExceptionWrapper]:
    if isinstance(exceptions, str):
        exceptions = Error(exceptions)
    elif isinstance(exceptions, int):
        exceptions = Panic(PanicCodeEnum(exceptions))

    if isinstance(exceptions, (tuple, list)):
        tmp: List[Union[str, int, Exception, Type[Exception]]] = []
        for ex in exceptions:
            if isinstance(ex, str):
                tmp.append(Error(ex))
            elif isinstance(ex, int):
                tmp.append(Panic(PanicCodeEnum(ex)))
            else:
                tmp.append(ex)
        exceptions = tuple(tmp)
        types = tuple(type(x) if not inspect.isclass(x) else x for x in exceptions)
    else:
        types = type(exceptions) if not inspect.isclass(exceptions) else exceptions

    wrapper = ExceptionWrapper()

    try:
        yield wrapper
    except types as e:  # pyright: ignore reportGeneralTypeIssues
        wrapper.value = e

        if isinstance(exceptions, (tuple, list)):
            if any(
                (inspect.isclass(ex) and issubclass(type(e), ex)) or e == ex
                for ex in exceptions
            ):
                return
            raise AssertionError(f"Expected revert of type {exceptions} but got {e}")
        else:
            if not inspect.isclass(exceptions):
                assert e == exceptions, f"Expected {e} but got {exceptions}"


def on_revert(callback: Callable[[TransactionRevertedError], None]):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except TransactionRevertedError as e:
                callback(e)
                raise

        return wrapper

    return decorator



import json
from urllib.request import Request, urlopen

from woke.utils import get_package_version

from .abc import ProtocolAbc


class HttpProtocol(ProtocolAbc):
    _uri: str
    _timeout: float

    def __init__(self, uri: str, timeout: float):
        self._uri = uri
        self._timeout = timeout

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def send_recv(self, data: str):
        req = Request(
            self._uri,
            data.encode("utf-8"),
            headers={
                "Content-Type": "application/json",
                "User-Agent": f"woke/{get_package_version('woke')}",
            },
        )

        with urlopen(req, timeout=self._timeout) as response:
            return json.loads(response.read().decode("utf-8"))



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from woke.ir.ast import SolcYulIdentifier

from ..utils import IrInitTuple
from .abc import YulAbc

if TYPE_CHECKING:
    from ..statements.inline_assembly import ExternalReference
    from .assignment import YulAssignment
    from .expression_statement import YulExpressionStatement
    from .for_loop import YulForLoop
    from .function_call import YulFunctionCall
    from .if_statement import YulIf
    from .switch import YulSwitch
    from .variable_declaration import YulVariableDeclaration


class YulIdentifier(YulAbc):
    """
    TBD
    """

    _parent: Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
    ]
    _name: str
    _external_reference: Optional[ExternalReference]

    def __init__(
        self, init: IrInitTuple, identifier: SolcYulIdentifier, parent: YulAbc
    ):
        super().__init__(init, identifier, parent)
        self._name = identifier.name
        self._external_reference = None

    @property
    def parent(
        self,
    ) -> Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
    ]:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def external_reference(self) -> Optional[ExternalReference]:
        return self._external_reference



from functools import lru_cache
from typing import Iterator, Optional, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcIndexAccess
from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.utils import IrInitTuple


class IndexAccess(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcIndexAccess
    _parent: SolidityAbc  # TODO: make this more specific

    _base_expression: ExpressionAbc
    _index_expression: Optional[ExpressionAbc]

    def __init__(
        self, init: IrInitTuple, index_access: SolcIndexAccess, parent: SolidityAbc
    ):
        super().__init__(init, index_access, parent)
        self._base_expression = ExpressionAbc.from_ast(
            init, index_access.base_expression, self
        )

        if index_access.index_expression is None:
            self._index_expression = None
        else:
            self._index_expression = ExpressionAbc.from_ast(
                init, index_access.index_expression, self
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._base_expression
        if self._index_expression is not None:
            yield from self._index_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def base_expression(self) -> ExpressionAbc:
        return self._base_expression

    @property
    def index_expression(self) -> Optional[ExpressionAbc]:
        return self._index_expression

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        return self.base_expression.is_ref_to_state_variable

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.base_expression.modifies_state
        if self.index_expression is not None:
            ret |= self.index_expression.modifies_state
        return ret



from typing import List, Optional, TypeVar

from typing_extensions import Annotated


class ValueRange:
    _min: int
    _max: int

    def __init__(self, min: int, max: int):
        self._min = min
        self._max = max

    @property
    def min(self) -> int:
        return self._min

    @property
    def max(self) -> int:
        return self._max


class Length:
    _length: int

    def __init__(self, length: int):
        self._length = length

    @property
    def length(self) -> int:
        return self._length


NoneType = type(None)


uint8 = Annotated[int, ValueRange(0, 2**8 - 1)]
uint16 = Annotated[int, ValueRange(0, 2**16 - 1)]
uint24 = Annotated[int, ValueRange(0, 2**24 - 1)]
uint32 = Annotated[int, ValueRange(0, 2**32 - 1)]
uint40 = Annotated[int, ValueRange(0, 2**40 - 1)]
uint48 = Annotated[int, ValueRange(0, 2**48 - 1)]
uint56 = Annotated[int, ValueRange(0, 2**56 - 1)]
uint64 = Annotated[int, ValueRange(0, 2**64 - 1)]
uint72 = Annotated[int, ValueRange(0, 2**72 - 1)]
uint80 = Annotated[int, ValueRange(0, 2**80 - 1)]
uint88 = Annotated[int, ValueRange(0, 2**88 - 1)]
uint96 = Annotated[int, ValueRange(0, 2**96 - 1)]
uint104 = Annotated[int, ValueRange(0, 2**104 - 1)]
uint112 = Annotated[int, ValueRange(0, 2**112 - 1)]
uint120 = Annotated[int, ValueRange(0, 2**120 - 1)]
uint128 = Annotated[int, ValueRange(0, 2**128 - 1)]
uint136 = Annotated[int, ValueRange(0, 2**136 - 1)]
uint144 = Annotated[int, ValueRange(0, 2**144 - 1)]
uint152 = Annotated[int, ValueRange(0, 2**152 - 1)]
uint160 = Annotated[int, ValueRange(0, 2**160 - 1)]
uint168 = Annotated[int, ValueRange(0, 2**168 - 1)]
uint176 = Annotated[int, ValueRange(0, 2**176 - 1)]
uint184 = Annotated[int, ValueRange(0, 2**184 - 1)]
uint192 = Annotated[int, ValueRange(0, 2**192 - 1)]
uint200 = Annotated[int, ValueRange(0, 2**200 - 1)]
uint208 = Annotated[int, ValueRange(0, 2**208 - 1)]
uint216 = Annotated[int, ValueRange(0, 2**216 - 1)]
uint224 = Annotated[int, ValueRange(0, 2**224 - 1)]
uint232 = Annotated[int, ValueRange(0, 2**232 - 1)]
uint240 = Annotated[int, ValueRange(0, 2**240 - 1)]
uint248 = Annotated[int, ValueRange(0, 2**248 - 1)]
uint256 = Annotated[int, ValueRange(0, 2**256 - 1)]
uint = uint256

int8 = Annotated[int, ValueRange(-(2**7), 2**7 - 1)]
int16 = Annotated[int, ValueRange(-(2**15), 2**15 - 1)]
int24 = Annotated[int, ValueRange(-(2**23), 2**23 - 1)]
int32 = Annotated[int, ValueRange(-(2**31), 2**31 - 1)]
int40 = Annotated[int, ValueRange(-(2**39), 2**39 - 1)]
int48 = Annotated[int, ValueRange(-(2**47), 2**47 - 1)]
int56 = Annotated[int, ValueRange(-(2**55), 2**55 - 1)]
int64 = Annotated[int, ValueRange(-(2**63), 2**63 - 1)]
int72 = Annotated[int, ValueRange(-(2**71), 2**71 - 1)]
int80 = Annotated[int, ValueRange(-(2**79), 2**79 - 1)]
int88 = Annotated[int, ValueRange(-(2**87), 2**87 - 1)]
int96 = Annotated[int, ValueRange(-(2**95), 2**95 - 1)]
int104 = Annotated[int, ValueRange(-(2**103), 2**103 - 1)]
int112 = Annotated[int, ValueRange(-(2**111), 2**111 - 1)]
int120 = Annotated[int, ValueRange(-(2**119), 2**119 - 1)]
int128 = Annotated[int, ValueRange(-(2**127), 2**127 - 1)]
int136 = Annotated[int, ValueRange(-(2**135), 2**135 - 1)]
int144 = Annotated[int, ValueRange(-(2**143), 2**143 - 1)]
int152 = Annotated[int, ValueRange(-(2**151), 2**151 - 1)]
int160 = Annotated[int, ValueRange(-(2**159), 2**159 - 1)]
int168 = Annotated[int, ValueRange(-(2**167), 2**167 - 1)]
int176 = Annotated[int, ValueRange(-(2**175), 2**175 - 1)]
int184 = Annotated[int, ValueRange(-(2**183), 2**183 - 1)]
int192 = Annotated[int, ValueRange(-(2**191), 2**191 - 1)]
int200 = Annotated[int, ValueRange(-(2**199), 2**199 - 1)]
int208 = Annotated[int, ValueRange(-(2**207), 2**207 - 1)]
int216 = Annotated[int, ValueRange(-(2**215), 2**215 - 1)]
int224 = Annotated[int, ValueRange(-(2**223), 2**223 - 1)]
int232 = Annotated[int, ValueRange(-(2**231), 2**231 - 1)]
int240 = Annotated[int, ValueRange(-(2**239), 2**239 - 1)]
int248 = Annotated[int, ValueRange(-(2**247), 2**247 - 1)]
int256 = Annotated[int, ValueRange(-(2**255), 2**255 - 1)]


bytes1 = Annotated[bytes, Length(1)]
bytes2 = Annotated[bytes, Length(2)]
bytes3 = Annotated[bytes, Length(3)]
bytes4 = Annotated[bytes, Length(4)]
bytes5 = Annotated[bytes, Length(5)]
bytes6 = Annotated[bytes, Length(6)]
bytes7 = Annotated[bytes, Length(7)]
bytes8 = Annotated[bytes, Length(8)]
bytes9 = Annotated[bytes, Length(9)]
bytes10 = Annotated[bytes, Length(10)]
bytes11 = Annotated[bytes, Length(11)]
bytes12 = Annotated[bytes, Length(12)]
bytes13 = Annotated[bytes, Length(13)]
bytes14 = Annotated[bytes, Length(14)]
bytes15 = Annotated[bytes, Length(15)]
bytes16 = Annotated[bytes, Length(16)]
bytes17 = Annotated[bytes, Length(17)]
bytes18 = Annotated[bytes, Length(18)]
bytes19 = Annotated[bytes, Length(19)]
bytes20 = Annotated[bytes, Length(20)]
bytes21 = Annotated[bytes, Length(21)]
bytes22 = Annotated[bytes, Length(22)]
bytes23 = Annotated[bytes, Length(23)]
bytes24 = Annotated[bytes, Length(24)]
bytes25 = Annotated[bytes, Length(25)]
bytes26 = Annotated[bytes, Length(26)]
bytes27 = Annotated[bytes, Length(27)]
bytes28 = Annotated[bytes, Length(28)]
bytes29 = Annotated[bytes, Length(29)]
bytes30 = Annotated[bytes, Length(30)]
bytes31 = Annotated[bytes, Length(31)]
bytes32 = Annotated[bytes, Length(32)]


T = TypeVar("T")
List1 = Annotated[List[T], Length(1)]
List2 = Annotated[List[T], Length(2)]
List3 = Annotated[List[T], Length(3)]
List4 = Annotated[List[T], Length(4)]
List5 = Annotated[List[T], Length(5)]
List6 = Annotated[List[T], Length(6)]
List7 = Annotated[List[T], Length(7)]
List8 = Annotated[List[T], Length(8)]
List9 = Annotated[List[T], Length(9)]
List10 = Annotated[List[T], Length(10)]
List11 = Annotated[List[T], Length(11)]
List12 = Annotated[List[T], Length(12)]
List13 = Annotated[List[T], Length(13)]
List14 = Annotated[List[T], Length(14)]
List15 = Annotated[List[T], Length(15)]
List16 = Annotated[List[T], Length(16)]
List17 = Annotated[List[T], Length(17)]
List18 = Annotated[List[T], Length(18)]
List19 = Annotated[List[T], Length(19)]
List20 = Annotated[List[T], Length(20)]
List21 = Annotated[List[T], Length(21)]
List22 = Annotated[List[T], Length(22)]
List23 = Annotated[List[T], Length(23)]
List24 = Annotated[List[T], Length(24)]
List25 = Annotated[List[T], Length(25)]
List26 = Annotated[List[T], Length(26)]
List27 = Annotated[List[T], Length(27)]
List28 = Annotated[List[T], Length(28)]
List29 = Annotated[List[T], Length(29)]
List30 = Annotated[List[T], Length(30)]
List31 = Annotated[List[T], Length(31)]
List32 = Annotated[List[T], Length(32)]



from __future__ import annotations

import re
from collections import deque
from functools import lru_cache, partial
from typing import (
    TYPE_CHECKING,
    Deque,
    FrozenSet,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from ..meta.modifier_invocation import ModifierInvocation
from ..meta.override_specifier import OverrideSpecifier
from ..reference_resolver import CallbackParams
from ..statements.block import Block
from .abc import DeclarationAbc

if TYPE_CHECKING:
    from woke.analysis.cfg import ControlFlowGraph
    from .contract_definition import ContractDefinition
    from .variable_declaration import VariableDeclaration
    from ..expressions.identifier import Identifier
    from ..expressions.member_access import MemberAccess
    from ..meta.identifier_path import IdentifierPathPart
    from ..meta.source_unit import SourceUnit
    from ..statements.inline_assembly import ExternalReference
    from ..expressions.unary_operation import UnaryOperation
    from ..expressions.binary_operation import BinaryOperation

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import AstNodeId, SolcFunctionDefinition, SolcStructuredDocumentation
from woke.ir.enums import FunctionKind, StateMutability, Visibility
from woke.ir.meta.parameter_list import ParameterList
from woke.ir.meta.structured_documentation import StructuredDocumentation
from woke.ir.utils import IrInitTuple


class FunctionDefinition(DeclarationAbc):
    """
    Definition of a function.

    !!! example
        Free function (= outside of a contract):
        ```solidity linenums="1"
        function f(uint a, uint b) pure returns (uint) {
            return a + b;
        }
        ```

        Function inside a contract (lines 2-4):
        ```solidity linenums="1"
        contract C {
            function f(uint a, uint b) public pure returns (uint) {
                return a + b;
            }
        }
        ```

        Constructor (lines 3-5):
        ```solidity linenums="1"
        contract C {
            uint public x;
            constructor(uint a) public {
                x = a;
            }
        }
        ```

        Fallback function (line 2):
        ```solidity linenums="1"
        contract C {
            fallback() external payable {}
        }
        ```

        Receive function (line 2):
        ```solidity linenums="1"
        contract C {
            receive() external payable {}
        }
        ```
    """

    _ast_node: SolcFunctionDefinition
    _parent: Union[ContractDefinition, SourceUnit]
    _child_functions: Set[Union[FunctionDefinition, VariableDeclaration]]

    _implemented: bool
    _kind: FunctionKind
    _modifiers: List[ModifierInvocation]
    _parameters: ParameterList
    _return_parameters: ParameterList
    # __scope
    _state_mutability: StateMutability
    _virtual: bool
    _visibility: Visibility
    _base_functions: List[AstNodeId]
    _documentation: Optional[Union[StructuredDocumentation, str]]
    _function_selector: Optional[bytes]
    _body: Optional[Block]
    _overrides: Optional[OverrideSpecifier]

    def __init__(
        self, init: IrInitTuple, function: SolcFunctionDefinition, parent: SolidityAbc
    ):
        super().__init__(init, function, parent)
        self._child_functions = set()

        self._implemented = function.implemented
        self._kind = function.kind

        if self._kind == FunctionKind.CONSTRUCTOR:
            self._name = "constructor"
        elif self._kind == FunctionKind.FALLBACK:
            self._name = "fallback"
        elif self._kind == FunctionKind.RECEIVE:
            self._name = "receive"

        self._modifiers = [
            ModifierInvocation(init, modifier, self) for modifier in function.modifiers
        ]
        self._parameters = ParameterList(init, function.parameters, self)
        self._return_parameters = ParameterList(init, function.return_parameters, self)
        # self.__scope = function.scope
        self._state_mutability = function.state_mutability
        self._virtual = function.virtual
        self._visibility = function.visibility
        self._base_functions = (
            list(function.base_functions) if function.base_functions is not None else []
        )
        if function.documentation is None:
            self._documentation = None
        elif isinstance(function.documentation, SolcStructuredDocumentation):
            self._documentation = StructuredDocumentation(
                init, function.documentation, self
            )
        elif isinstance(function.documentation, str):
            self._documentation = function.documentation
        else:
            raise TypeError(
                f"Unknown type of documentation: {type(function.documentation)}"
            )
        self._function_selector = (
            bytes.fromhex(function.function_selector)
            if function.function_selector
            else None
        )

        if (
            self._visibility in {Visibility.PUBLIC, Visibility.EXTERNAL}
            and self._kind == FunctionKind.FUNCTION
        ):
            assert self._function_selector is not None
        else:
            assert self._function_selector is None

        self._body = Block(init, function.body, self) if function.body else None
        assert (self._body is not None) == self._implemented
        self._overrides = (
            OverrideSpecifier(init, function.overrides, self)
            if function.overrides
            else None
        )
        self._reference_resolver.register_post_process_callback(self._post_process)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for modifier in self._modifiers:
            yield from modifier
        yield from self._parameters
        yield from self._return_parameters
        if isinstance(self._documentation, StructuredDocumentation):
            yield from self._documentation
        if self._body is not None:
            yield from self._body
        if self._overrides is not None:
            yield from self._overrides

    def _post_process(self, callback_params: CallbackParams):
        base_functions = self.base_functions
        for base_function in base_functions:
            base_function._child_functions.add(self)
        self._reference_resolver.register_destroy_callback(
            self.file, partial(self._destroy, base_functions)
        )

    def _destroy(self, base_functions: Tuple[FunctionDefinition]) -> None:
        for base_function in base_functions:
            base_function._child_functions.remove(self)

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        FUNCTION_RE = re.compile(
            r"^\s*function\s+(?P<name>{identifier})".format(
                identifier=IDENTIFIER
            ).encode("utf-8")
        )
        CONSTRUCTOR_RE = re.compile(r"^\s*(?P<name>constructor)".encode("utf-8"))
        FALLBACK_RE = re.compile(r"^\s*(?P<name>fallback)".encode("utf-8"))
        RECEIVE_RE = re.compile(r"^\s*(?P<name>receive)".encode("utf-8"))

        regexps = [FUNCTION_RE, CONSTRUCTOR_RE, FALLBACK_RE, RECEIVE_RE]
        matches = [regexp.match(self._source) for regexp in regexps]
        assert any(matches)

        byte_start = self._ast_node.src.byte_offset
        match = next(match for match in matches if match)
        return byte_start + match.start("name"), byte_start + match.end("name")

    def get_all_references(
        self, include_declarations: bool
    ) -> Iterator[
        Union[
            DeclarationAbc,
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ]
    ]:
        from .variable_declaration import VariableDeclaration

        processed_declarations: Set[Union[FunctionDefinition, VariableDeclaration]] = {
            self
        }
        declarations_queue: Deque[
            Union[FunctionDefinition, VariableDeclaration]
        ] = deque([self])

        while declarations_queue:
            declaration = declarations_queue.pop()
            if include_declarations:
                yield declaration
            yield from declaration.references

            if isinstance(declaration, (FunctionDefinition, VariableDeclaration)):
                for base_function in declaration.base_functions:
                    if base_function not in processed_declarations:
                        declarations_queue.append(base_function)
                        processed_declarations.add(base_function)
            if isinstance(declaration, FunctionDefinition):
                for child_function in declaration.child_functions:
                    if child_function not in processed_declarations:
                        declarations_queue.append(child_function)
                        processed_declarations.add(child_function)

    @property
    def parent(self) -> Union[ContractDefinition, SourceUnit]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    @lru_cache(maxsize=2048)
    def canonical_name(self) -> str:
        from .contract_definition import ContractDefinition

        if isinstance(self._parent, ContractDefinition):
            return f"{self._parent.canonical_name}.{self._name}({','.join(param.type_name.type_string for param in self._parameters.parameters)})"
        return f"{self._name}({','.join(param.type_name.type_string for param in self._parameters.parameters)})"

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        if self.kind == FunctionKind.CONSTRUCTOR:
            ret = "constructor"
        elif self.kind == FunctionKind.FALLBACK:
            ret = "fallback"
        elif self.kind == FunctionKind.RECEIVE:
            ret = "receive"
        else:
            ret = f"function {self.name}"
        ret += f"({', '.join(parameter.declaration_string for parameter in self.parameters.parameters)})"
        ret += f" {self.visibility}"
        ret += (
            f" {self.state_mutability}"
            if self.state_mutability != StateMutability.NONPAYABLE
            else ""
        )
        ret += f" virtual" if self.virtual else ""
        ret += (
            (
                f" override"
                + (
                    "("
                    + ", ".join(
                        override.source for override in self.overrides.overrides
                    )
                    + ")"
                    if len(self.overrides.overrides) > 0
                    else ""
                )
            )
            if self.overrides is not None
            else ""
        )
        ret += (
            (" " + " ".join(modifier.source for modifier in self.modifiers))
            if len(self.modifiers) > 0
            else ""
        )
        ret += (
            " returns ("
            + ", ".join(
                parameter.declaration_string
                for parameter in self.return_parameters.parameters
            )
            + ")"
            if len(self.return_parameters.parameters) > 0
            else ""
        )

        if isinstance(self.documentation, StructuredDocumentation):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.text.splitlines())
                + "\n"
                + ret
            )
        elif isinstance(self.documentation, str):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.splitlines())
                + "\n"
                + ret
            )
        else:
            return ret

    @property
    def implemented(self) -> bool:
        """
        Returns:
            `True` if the function [body][woke.ir.declarations.function_definition.FunctionDefinition.body] is not `None`, `False` otherwise.
        """
        return self._implemented

    @property
    def kind(self) -> FunctionKind:
        """
        Returns:
            Kind of the function.
        """
        return self._kind

    @property
    def modifiers(self) -> Tuple[ModifierInvocation, ...]:
        """
        Also includes base constructor invocations.
        !!! example
            Both `:::solidity ERC20Token("My Token", "MTK", msg.sender, 10 ** 18)` and `initializer` are listed by this property.
            ```solidity
            contract MyToken is ERC20Token {
                constructor() ERC20Token("My Token", "MTK", msg.sender, 10 ** 18) initializer {}
            }
            ```

        Returns:
            List of modifiers applied to the function.
        """
        return tuple(self._modifiers)

    @property
    def parameters(self) -> ParameterList:
        """
        Returns:
            Parameter list describing the function parameters.
        """
        return self._parameters

    @property
    def return_parameters(self) -> ParameterList:
        """
        Returns:
            Parameter list describing the function return parameters.
        """
        return self._return_parameters

    @property
    def state_mutability(self) -> StateMutability:
        """
        Returns:
            State mutability of the function.
        """
        return self._state_mutability

    @property
    def virtual(self) -> bool:
        """
        Returns:
            `True` if the function is virtual, `False` otherwise.
        """
        return self._virtual

    @property
    def visibility(self) -> Visibility:
        """
        Returns:
            Visibility of the function.
        """
        return self._visibility

    @property
    def base_functions(self) -> Tuple[FunctionDefinition, ...]:
        """
        !!! example
            `A.foo` on lines 6-8 lists `I.foo` on line 2 as a base function.

            `B.foo` on lines 12-14 lists only `A.foo` on lines 6-8 as a base function.
            ```solidity linenums="1"
            interface I {
                function foo() external returns(uint);
            }

            contract A is I {
                function foo() external pure virtual override returns(uint) {
                    return 1;
                }
            }

            contract B is A {
                function foo() external pure override returns(uint) {
                    return 2;
                }
            }
            ```

        !!! example
            `A1.foo` on lines 6-8 lists `I.foo` on line 2 as a base function.

            `A2.foo` on lines 12-14 lists `I.foo` on line 2 as a base function.

            `B.foo` on lines 18-20 lists `A1.foo` on lines 6-8 and `A2.foo` on lines 12-14 as base functions.
            ```solidity linenums="1"
            interface I {
                function foo() external returns(uint);
            }

            contract A1 is I {
                function foo() external pure virtual override returns(uint) {
                    return 1;
                }
            }

            contract A2 is I {
                function foo() external pure virtual override returns(uint) {
                    return 2;
                }
            }

            contract B is A1, A2 {
                function foo() external pure override(A1, A2) returns(uint) {
                    return 3;
                }
            }
            ```
        Returns:
            List of base functions overridden by this function.
        """
        base_functions = []
        for base_function_id in self._base_functions:
            base_function = self._reference_resolver.resolve_node(
                base_function_id, self._cu_hash
            )
            assert isinstance(base_function, FunctionDefinition)
            base_functions.append(base_function)
        return tuple(base_functions)

    @property
    def child_functions(
        self,
    ) -> FrozenSet[Union[FunctionDefinition, VariableDeclaration]]:
        """
        Returns:
            Functions that list this function in their [base_functions][woke.ir.declarations.function_definition.FunctionDefinition.base_functions] property.
        """
        return frozenset(self._child_functions)

    @property
    def documentation(self) -> Optional[Union[StructuredDocumentation, str]]:
        """
        Of [StructuredDocumentation][woke.ir.meta.structured_documentation.StructuredDocumentation] type since Solidity 0.6.3.
        Returns:
            [NatSpec](https://solidity.readthedocs.io/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation

    @property
    def function_selector(self) -> Optional[bytes]:
        """
        Is only set for [Visibility.PUBLIC][woke.ir.enums.Visibility.PUBLIC] and [Visibility.EXTERNAL][woke.ir.enums.Visibility.EXTERNAL] functions of the [FunctionKind.FUNCTION][woke.ir.enums.FunctionKind.FUNCTION] kind.
        Returns:
            Selector of the function.
        """
        return self._function_selector

    @property
    def body(self) -> Optional[Block]:
        """
        Returns:
            Body of the function, if any.
        """
        return self._body

    @property
    def overrides(self) -> Optional[OverrideSpecifier]:
        """
        Returns override specifier as present in the source code.
        !!! example
            `I.foo` on line 2 does not have an override specifier.

            `A.foo` on lines 6-8 has an override specifier with the [overrides][woke.ir.meta.override_specifier.OverrideSpecifier.overrides] property empty.

            `B.foo` on lines 12-14 has an override specifier with the [overrides][woke.ir.meta.override_specifier.OverrideSpecifier.overrides] property containg one item referencing the contract `A` ([ContractDefinition][woke.ir.declarations.contract_definition.ContractDefinition]).
            ```solidity linenums="1"
            interface I {
                function foo() external returns(uint);
            }

            contract A is I {
                function foo() external pure virtual override returns(uint) {
                    return 1;
                }
            }

            contract B is A {
                function foo() external pure override(A) returns(uint) {
                    return 2;
                }
            }
            ```

        Returns:
            Override specifier, if any.
        """
        return self._overrides

    @property
    @lru_cache(maxsize=2048)
    def cfg(self) -> Optional[ControlFlowGraph]:
        from woke.analysis.cfg import ControlFlowGraph

        if self.body is None:
            return None
        return ControlFlowGraph(self)



from __future__ import annotations

import asyncio
import logging
import os
import pickle
import time
from collections import defaultdict, deque
from contextlib import nullcontext
from json import JSONDecodeError
from pathlib import Path
from types import MappingProxyType
from typing import (
    Callable,
    Collection,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
)

import networkx as nx
import rich
import rich.console
import rich.panel
from Crypto.Hash import BLAKE2b
from intervaltree import IntervalTree
from pathvalidate import sanitize_filename  # type: ignore
from pydantic import ValidationError
from rich.progress import Progress
from watchdog.events import (
    FileSystemEvent,
    FileSystemEventHandler,
    FileSystemMovedEvent,
)

from woke.config import WokeConfig
from woke.core.solidity_version import (
    SolidityVersion,
    SolidityVersionRange,
    SolidityVersionRanges,
)
from woke.ir import SourceUnit
from woke.ir.ast import AstSolc
from woke.ir.reference_resolver import CallbackParams, ReferenceResolver
from woke.ir.utils import IrInitTuple
from woke.regex_parser import SoliditySourceParser
from woke.svm import SolcVersionManager

from ..utils import get_package_version
from ..utils.file_utils import is_relative_to
from .build_data_model import (
    CompilationUnitBuildInfo,
    ProjectBuild,
    ProjectBuildInfo,
    SourceUnitInfo,
)
from .compilation_unit import CompilationUnit
from .exceptions import CompilationError, CompilationResolveError
from .solc_frontend import (
    SolcFrontend,
    SolcInputOptimizerSettings,
    SolcInputSettings,
    SolcOutput,
    SolcOutputError,
    SolcOutputErrorSeverityEnum,
    SolcOutputSelectionEnum,
)
from .source_path_resolver import SourcePathResolver
from .source_unit_name_resolver import SourceUnitNameResolver

# pyright: reportGeneralTypeIssues=false

logger = logging.getLogger(__name__)


class CompilationFileSystemEventHandler(FileSystemEventHandler):
    _config: WokeConfig
    _files: Set[Path]
    _config_path: Path
    _compiler: SolidityCompiler
    _output_types: Collection[SolcOutputSelectionEnum]
    _write_artifacts: bool
    _console: Optional[rich.console.Console]
    _no_warnings: bool
    _created_files: Set[Path]
    _modified_files: Set[Path]
    _deleted_files: Set[Path]
    _config_changed: bool
    _loop: asyncio.AbstractEventLoop
    _queue: asyncio.Queue[FileSystemEvent]
    _callbacks: List[Callable[[ProjectBuild, ProjectBuildInfo], None]]

    TIMEOUT_INTERVAL = 1.0

    def __init__(
        self,
        config: WokeConfig,
        files: Iterable[Path],
        loop: asyncio.AbstractEventLoop,
        compiler: SolidityCompiler,
        output_types: Collection[SolcOutputSelectionEnum],
        *,
        write_artifacts: bool = True,
        console: Optional[rich.console.Console] = None,
        no_warnings: bool = False,
    ):
        self._config = config
        self._files = set(files)
        self._loop = loop
        self._config_path = config.project_root_path / "woke.toml"
        self._compiler = compiler
        self._output_types = output_types
        self._write_artifacts = write_artifacts
        self._console = console
        self._no_warnings = no_warnings
        self._created_files = set()
        self._modified_files = set()
        self._deleted_files = set()
        self._config_changed = False
        self._queue = asyncio.Queue()
        self._callbacks = []

    async def run(self):
        while True:
            # process at least one event
            event = await self._queue.get()
            self._process_event(event)

            start = time.perf_counter()
            while time.perf_counter() - start < self.TIMEOUT_INTERVAL:
                try:
                    event = self._queue.get_nowait()
                    self._process_event(event)
                except asyncio.QueueEmpty:
                    await asyncio.sleep(0.1)

            await self._compile()

            assert self._compiler.latest_build is not None
            assert self._compiler.latest_build_info is not None

            for callback in self._callbacks:
                callback(self._compiler.latest_build, self._compiler.latest_build_info)

    def register_callback(
        self, callback: Callable[[ProjectBuild, ProjectBuildInfo], None]
    ):
        self._callbacks.append(callback)

    def unregister_callback(
        self, callback: Callable[[ProjectBuild, ProjectBuildInfo], None]
    ):
        self._callbacks.remove(callback)

    def _process_event(self, event: FileSystemEvent):
        if isinstance(event, FileSystemMovedEvent):
            self._on_deleted(Path(event.src_path))
            self._on_created(Path(event.dest_path))
        elif event.event_type == "created":
            self._on_created(Path(event.src_path))
        elif event.event_type == "modified":
            self._on_modified(Path(event.src_path))
        elif event.event_type == "deleted":
            self._on_deleted(Path(event.src_path))

    def on_any_event(self, event: FileSystemEvent):
        if event.is_directory:
            return
        if isinstance(event, FileSystemMovedEvent):
            src_file = Path(event.src_path)
            dest_file = Path(event.dest_path)
            if (
                src_file == self._config_path
                or src_file.suffix == ".sol"
                or dest_file == self._config_path
                or dest_file.suffix == ".sol"
            ):
                self._loop.call_soon_threadsafe(self._queue.put_nowait, event)
        else:
            file = Path(event.src_path)
            if file == self._config_path or file.suffix == ".sol":
                self._loop.call_soon_threadsafe(self._queue.put_nowait, event)

    async def _compile(self):
        if self._config_changed:
            self._config.load_configs()

            # find files that were previously ignored but are now included
            ctx_manager = (
                self._console.status("[bold green]Searching for *.sol files...[/]")
                if self._console
                else nullcontext()
            )
            start = time.perf_counter()
            files = set()
            with ctx_manager:
                for file in self._config.project_root_path.rglob("**/*.sol"):
                    if (
                        not any(
                            is_relative_to(file, p)
                            for p in self._config.compiler.solc.ignore_paths
                        )
                        and file.is_file()
                    ):
                        files.add(file)
            end = time.perf_counter()
            if self._console is not None:
                self._console.log(
                    f"[green]Found {len(files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
                )

            deleted_files = self._deleted_files
        else:
            files = self._files.copy()
            files.update(self._created_files)
            files.update(self._modified_files)
            ignored_files = {
                f
                for f in files
                if any(
                    is_relative_to(f, p)
                    for p in self._config.compiler.solc.ignore_paths
                )
                or not is_relative_to(f, self._config.project_root_path)
            }
            files.difference_update(ignored_files)
            files.difference_update(self._deleted_files)
            deleted_files = self._deleted_files

        await self._compiler.compile(
            files,
            self._output_types,
            write_artifacts=self._write_artifacts,
            deleted_files=deleted_files,
            console=self._console,
            no_warnings=self._no_warnings,
        )

        self._created_files.clear()
        self._modified_files.clear()
        self._deleted_files.clear()
        self._config_changed = False

    def _on_created(self, file: Path):
        if file == self._config_path:
            self._config_changed = True
        elif file.suffix == ".sol":
            self._files.add(file)
            if file in self._deleted_files:
                self._deleted_files.remove(file)
                self._modified_files.add(file)
            else:
                logger.debug(f"File {file} created")
                self._created_files.add(file)

    def _on_modified(self, file: Path):
        if file == self._config_path:
            self._config_changed = True
        elif file.suffix == ".sol":
            logger.debug(f"File {file} modified")
            self._modified_files.add(file)

    def _on_deleted(self, file: Path):
        if file == self._config_path:
            self._config_changed = True
        elif file.suffix == ".sol":
            try:
                self._files.remove(file)
            except KeyError:
                pass
            if file in self._modified_files:
                self._modified_files.remove(file)

            if file in self._created_files:
                self._created_files.remove(file)
            else:
                logger.debug(f"File {file} deleted")
                self._deleted_files.add(file)


class SolidityCompiler:
    __config: WokeConfig
    __svm: SolcVersionManager
    __solc_frontend: SolcFrontend
    __source_unit_name_resolver: SourceUnitNameResolver
    __source_path_resolver: SourcePathResolver

    _latest_build_info: Optional[ProjectBuildInfo]
    _latest_build: Optional[ProjectBuild]
    _latest_graph: Optional[nx.DiGraph]
    _latest_source_units_to_paths: Optional[Dict[str, Path]]

    def __init__(self, woke_config: WokeConfig):
        self.__config = woke_config
        self.__svm = SolcVersionManager(woke_config)
        self.__solc_frontend = SolcFrontend(woke_config)
        self.__source_unit_name_resolver = SourceUnitNameResolver(woke_config)
        self.__source_path_resolver = SourcePathResolver(woke_config)

        self._latest_build_info = None
        self._latest_build = None
        self._latest_graph = None
        self._latest_source_units_to_paths = None

    @property
    def latest_build_info(self) -> Optional[ProjectBuildInfo]:
        return self._latest_build_info

    @property
    def latest_build(self) -> Optional[ProjectBuild]:
        return self._latest_build

    @property
    def latest_graph(self) -> Optional[nx.DiGraph]:
        return self._latest_graph

    @property
    def latest_source_units_to_paths(self) -> Optional[MappingProxyType[str, Path]]:
        return MappingProxyType(self._latest_source_units_to_paths)

    def build_graph(
        self,
        files: Iterable[Path],
        modified_files: Mapping[Path, str],
        ignore_errors: bool = False,
    ) -> Tuple[nx.DiGraph, Dict[str, Path]]:
        # source unit name, full path, file content
        source_units_queue: deque[Tuple[str, Path, Optional[str]]] = deque()
        source_units: Dict[str, Path] = {}

        # for every source file resolve a source unit name
        for file in files:
            try:
                file = file.resolve(strict=True)
            except FileNotFoundError:
                if file in modified_files:
                    pass
                elif ignore_errors:
                    continue
                else:
                    raise

            source_unit_name = self.__source_unit_name_resolver.resolve_cmdline_arg(
                str(file)
            )
            if source_unit_name in source_units:
                first = str(source_units[source_unit_name])
                second = str(file)
                raise CompilationError(
                    f"Same source unit name `{source_unit_name}` for multiple source files:\n{first}\n{second}"
                )
            source_units[source_unit_name] = file
            content = modified_files.get(file, None)
            source_units_queue.append((source_unit_name, file, content))

        graph = nx.DiGraph()

        # recursively process all sources
        while len(source_units_queue) > 0:
            source_unit_name, path, content = source_units_queue.pop()
            if content is None:
                try:
                    versions, imports, h, content = SoliditySourceParser.parse(
                        path, ignore_errors
                    )
                except UnicodeDecodeError:
                    continue
            else:
                versions, imports, h = SoliditySourceParser.parse_source(
                    content, ignore_errors
                )
            graph.add_node(
                source_unit_name,
                path=path,
                versions=versions,
                hash=h,
                content=content,
                unresolved_imports=set(),
            )
            source_units[source_unit_name] = path

            for _import in imports:
                import_unit_name = self.__source_unit_name_resolver.resolve_import(
                    source_unit_name, _import
                )
                try:
                    import_path = self.__source_path_resolver.resolve(
                        import_unit_name, source_unit_name, modified_files.keys()
                    ).resolve()
                    if import_path not in modified_files:
                        import_path = import_path.resolve(strict=True)
                except (FileNotFoundError, CompilationResolveError):
                    if ignore_errors:
                        graph.nodes[source_unit_name]["unresolved_imports"].add(
                            import_unit_name
                        )
                        continue
                    raise

                if import_unit_name in source_units:
                    other_path = source_units[import_unit_name]
                    if import_path != other_path:
                        raise ValueError(
                            f"Same source unit name `{import_unit_name}` for multiple source files:\n{import_path}\n{other_path}"
                        )
                else:
                    source_units_queue.append(
                        (
                            import_unit_name,
                            import_path,
                            modified_files.get(import_path, None),
                        )
                    )
                graph.add_edge(import_unit_name, source_unit_name)
        return graph, source_units

    @staticmethod
    def build_compilation_units_maximize(graph: nx.DiGraph) -> List[CompilationUnit]:
        """
        Builds a list of compilation units from a graph. Number of compilation units is maximized.
        """

        def __build_compilation_unit(
            graph: nx.DiGraph, start: Iterable[str]
        ) -> CompilationUnit:
            nodes_subset = set()
            nodes_queue: deque[str] = deque(start)
            versions: SolidityVersionRanges = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )

            while len(nodes_queue) > 0:
                node = nodes_queue.pop()
                versions &= graph.nodes[node]["versions"]

                if node in nodes_subset:
                    continue
                nodes_subset.add(node)

                for in_edge in graph.in_edges(node):
                    _from, to = in_edge
                    if _from not in nodes_subset:
                        nodes_queue.append(_from)

            subgraph = graph.subgraph(nodes_subset).copy()
            return CompilationUnit(subgraph, versions)

        sinks = [node for node, out_degree in graph.out_degree() if out_degree == 0]
        compilation_units = []

        for sink in sinks:
            compilation_unit = __build_compilation_unit(graph, [sink])
            compilation_units.append(compilation_unit)

        # cycles can also be "sinks" in terms of compilation units
        generated_cycles: Set[FrozenSet[str]] = set()

        for cycle in nx.simple_cycles(graph):
            if frozenset(cycle) in generated_cycles:
                continue

            is_closed_cycle = True
            for node in cycle:
                if any(edge[1] not in cycle for edge in graph.out_edges(node)):
                    is_closed_cycle = False
                    break

            if is_closed_cycle:
                generated_cycles.add(frozenset(cycle))
                compilation_unit = __build_compilation_unit(graph, cycle)
                compilation_units.append(compilation_unit)
        return compilation_units

    def create_build_settings(
        self, output_types: Collection[SolcOutputSelectionEnum]
    ) -> SolcInputSettings:
        settings = SolcInputSettings()  # type: ignore
        # TODO Allow setting all solc build settings
        # Currently it is not possible to set all solc standard JSON input build settings.
        # These include: stopAfter, optimizer, via_IR, debug, metadata, libraries and model checker settings.
        # See https://docs.soliditylang.org/en/v0.8.12/using-the-compiler.html#input-description.
        # Also it is not possible to specify solc output per contract or per source file.
        settings.remappings = [
            str(remapping) for remapping in self.__config.compiler.solc.remappings
        ]
        settings.evm_version = self.__config.compiler.solc.evm_version
        settings.via_IR = self.__config.compiler.solc.via_IR
        settings.optimizer = SolcInputOptimizerSettings(
            enabled=self.__config.compiler.solc.optimizer.enabled,
            runs=self.__config.compiler.solc.optimizer.runs,
        )

        settings.output_selection = {"*": {}}

        if SolcOutputSelectionEnum.ALL in output_types:
            settings.output_selection["*"][""] = [SolcOutputSelectionEnum.AST]  # type: ignore
            settings.output_selection["*"]["*"] = [SolcOutputSelectionEnum.ALL]  # type: ignore
        else:
            if SolcOutputSelectionEnum.AST in output_types:
                settings.output_selection["*"][""] = [SolcOutputSelectionEnum.AST]  # type: ignore
            settings.output_selection["*"]["*"] = [
                output_type
                for output_type in output_types
                if output_type != SolcOutputSelectionEnum.AST
            ]  # type: ignore

        return settings

    def determine_solc_versions(
        self, compilation_units: Iterable[CompilationUnit]
    ) -> List[SolidityVersion]:
        target_versions = []
        min_version = self.__config.min_solidity_version
        max_version = self.__config.max_solidity_version
        for compilation_unit in compilation_units:
            target_version = self.__config.compiler.solc.target_version
            if target_version is not None:
                if target_version not in compilation_unit.versions:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"Unable to compile following files with solc version `{target_version}` set in config files:\n"
                        + files_str
                    )
            else:
                # use the latest matching version
                matching_versions = [
                    version
                    for version in reversed(self.__svm.list_all())
                    if version in compilation_unit.versions
                ]
                if len(matching_versions) == 0:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"Unable to compile following files with any solc version:\n"
                        + files_str
                    )
                try:
                    target_version = next(
                        version
                        for version in matching_versions
                        if version <= max_version
                    )
                except StopIteration:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"The maximum supported version of Solidity is {max_version}, unable to compile the following files:\n"
                        + files_str,
                    )
                if target_version < min_version:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"The minimum supported version of Solidity is {min_version}, unable to compile the following files:\n"
                        + files_str,
                    )
            target_versions.append(target_version)
        return target_versions

    async def _install_solc(
        self,
        versions: Iterable[SolidityVersion],
        console: Optional[rich.console.Console],
    ) -> None:
        for version in set(versions):
            if not self.__svm.installed(version):
                if console is None:
                    await self.__svm.install(version)
                else:
                    with Progress(console=console) as progress:
                        task = progress.add_task(
                            f"[green]Downloading solc {version}", total=1
                        )

                        async def on_progress(downloaded: int, total: int) -> None:
                            progress.update(task, completed=downloaded, total=total)  # type: ignore

                        await self.__svm.install(
                            version,
                            progress=on_progress,
                        )

    @staticmethod
    def _out_edge_bfs(
        cu: CompilationUnit, start: Iterable[Path], out: Set[Path]
    ) -> None:
        processed: Set[str] = set()
        for path in start:
            processed.update(cu.path_to_source_unit_names(path))
        out.update(start)

        queue: Deque[str] = deque(processed)
        while len(queue):
            node = queue.pop()
            for out_edge in cu.graph.out_edges(node):
                from_, to = out_edge
                if to not in processed:
                    processed.add(to)
                    queue.append(to)
                    out.add(cu.source_unit_name_to_path(to))

    def load(self, *, console: Optional[rich.console.Console] = None) -> None:
        ctx_manager = (
            console.status("[bold green]Loading previous build...")
            if console is not None
            else nullcontext()
        )
        start = time.perf_counter()

        with ctx_manager:
            try:
                latest_build_path = self.__config.project_root_path / ".woke-build"
                build_info = ProjectBuildInfo.parse_file(
                    latest_build_path / "build.json"
                )
                build_data = (latest_build_path / "build.bin").read_bytes()

                if build_info.woke_version != get_package_version("woke"):
                    if console is not None:
                        console.log(
                            f"[yellow]Woke version changed from {build_info.woke_version} to {get_package_version('woke')} since the last build[/yellow]"
                        )
                    return

                build_key_path = self.__config.global_data_path / "build.key"
                if not build_key_path.is_file():
                    if console is not None:
                        console.log(
                            f"[yellow]No build key found, cannot verify build signature.[/yellow]"
                        )
                build_key = build_key_path.read_bytes()
                if len(build_key) != 64:
                    if console is not None:
                        console.log(
                            f"[yellow]Loaded build key with incorrect length, cannot verify build signature.[/yellow]"
                        )
                    return

                build_sig = (latest_build_path / "build.bin.sig").read_bytes()

                h = BLAKE2b.new(digest_bits=512, key=build_key)
                h.update(build_data)
                h.verify(build_sig)

                self._latest_build = pickle.loads(build_data)
                self._latest_build_info = build_info
            except (
                AttributeError,
                ModuleNotFoundError,
                ValidationError,
                JSONDecodeError,
                FileNotFoundError,
                pickle.UnpicklingError,
                ValueError,
            ):
                if console is not None:
                    console.log("[red]Failed to load previous build artifacts[/red]")
                return

        if console is not None:
            end = time.perf_counter()
            console.log(
                f"[green]Loaded previous build in [bold green]{end - start:.2f} s[/bold green]"
            )

    @staticmethod
    def _merge_compilation_units(
        compilation_units: List[CompilationUnit],
        graph: nx.DiGraph,
    ) -> List[CompilationUnit]:
        # optimization - merge compilation units that can be compiled together
        if len(compilation_units) > 0 and all(
            len(cu.versions) for cu in compilation_units
        ):
            compilation_units = sorted(
                compilation_units, key=lambda cu: cu.versions.version_ranges[0].lower
            )

            merged_compilation_units: List[CompilationUnit] = []
            source_unit_names: Set = set()
            versions = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )

            for cu in compilation_units:
                if versions & cu.versions:
                    source_unit_names |= cu.source_unit_names
                    versions &= cu.versions
                else:
                    merged_compilation_units.append(
                        CompilationUnit(
                            graph.subgraph(source_unit_names).copy(),
                            versions,
                        )
                    )
                    source_unit_names = set(cu.source_unit_names)
                    versions = cu.versions

            merged_compilation_units.append(
                CompilationUnit(
                    graph.subgraph(source_unit_names).copy(),
                    versions,
                )
            )

            logger.debug(
                f"Merged {len(compilation_units)} compilation units into {len(merged_compilation_units)}"
            )
            compilation_units = merged_compilation_units

        return compilation_units

    async def compile(
        self,
        files: Iterable[Path],
        output_types: Collection[SolcOutputSelectionEnum],
        *,
        write_artifacts: bool = True,
        force_recompile: bool = False,
        modified_files: Optional[Mapping[Path, str]] = None,
        deleted_files: Optional[
            Set[Path]
        ] = None,  # files that should be treated as deleted even if they exist
        console: Optional[rich.console.Console] = None,
        no_warnings: bool = False,
        incremental: Optional[bool] = None,
    ) -> Tuple[ProjectBuild, Set[SolcOutputError]]:
        if modified_files is None:
            modified_files = {}
        if deleted_files is None:
            deleted_files = set()

        if incremental is None:
            if force_recompile or self._latest_build_info is None:
                incremental = True
            else:
                incremental = self._latest_build_info.incremental

        # validate target solc version (if set)
        target_version = self.__config.compiler.solc.target_version
        min_version = self.__config.min_solidity_version
        max_version = self.__config.max_solidity_version
        if target_version is not None and target_version < min_version:
            raise CompilationError(
                f"Target configured version {target_version} is lower than minimum supported version {min_version}"
            )
        if target_version is not None and target_version > max_version:
            raise CompilationError(
                f"Target configured version {target_version} is higher than maximum supported version {max_version}"
            )

        graph, source_units_to_paths = self.build_graph(
            files, modified_files, ignore_errors=True
        )
        compilation_units = self.build_compilation_units_maximize(graph)
        build_settings = self.create_build_settings(output_types)

        self._latest_graph = graph
        self._latest_source_units_to_paths = source_units_to_paths

        build_settings_changed = False
        if self._latest_build_info is not None:
            if (
                self._latest_build_info.allow_paths
                != self.__config.compiler.solc.allow_paths
                or self._latest_build_info.ignore_paths
                != self.__config.compiler.solc.ignore_paths
                or self._latest_build_info.include_paths
                != self.__config.compiler.solc.include_paths
                or self._latest_build_info.settings != build_settings
                or self._latest_build_info.target_solidity_version
                != self.__config.compiler.solc.target_version
                or self._latest_build_info.incremental != incremental
            ):
                logger.debug("Build settings changed")
                build_settings_changed = True

        errors_per_cu: DefaultDict[bytes, Set[SolcOutputError]] = defaultdict(set)

        if (
            force_recompile
            or self._latest_build_info is None
            or self._latest_build is None
            or build_settings_changed
        ):
            logger.debug("Performing full recompile")
            build = ProjectBuild(
                interval_trees={},
                reference_resolver=ReferenceResolver(),
                source_units={},
            )
            files_to_compile = set(
                source_units_to_paths[source_unit] for source_unit in graph.nodes
            )

            if not incremental:
                compilation_units = self._merge_compilation_units(
                    compilation_units, graph
                )
        else:
            # TODO this is not needed? graph contains hash of modified files
            # files_to_compile = set(modified_files.keys())
            files_to_compile = set()

            for source_unit in graph.nodes:
                if (
                    source_unit not in self._latest_build_info.source_units_info
                    or self._latest_build_info.source_units_info[
                        source_unit
                    ].blake2b_hash
                    != graph.nodes[source_unit]["hash"]
                ):
                    files_to_compile.add(source_units_to_paths[source_unit])

            for source_unit, info in self._latest_build_info.source_units_info.items():
                if source_unit not in graph.nodes:
                    deleted_files.add(info.fs_path)

            if not incremental:
                compilation_units = self._merge_compilation_units(
                    compilation_units, graph
                )

            for cu_hash, cu_data in self._latest_build_info.compilation_units.items():
                if any(cu.hash.hex() == cu_hash for cu in compilation_units):
                    errors_per_cu[bytes.fromhex(cu_hash)] = set(cu_data.errors)

            # select only compilation units that need to be compiled
            compilation_units = [
                cu
                for cu in compilation_units
                if (cu.files & files_to_compile)
                or cu.contains_unresolved_file(deleted_files, self.__config)
            ]

            logger.debug(
                f"Recompiling {len(files_to_compile)} files and {len(compilation_units)} compilation units"
            )

            build = self._latest_build

        target_versions = self.determine_solc_versions(compilation_units)
        await self._install_solc(target_versions, console)

        tasks = []
        for compilation_unit, target_version in zip(compilation_units, target_versions):
            task = asyncio.create_task(
                self.compile_unit_raw(compilation_unit, target_version, build_settings)
            )
            tasks.append(task)

        logger.debug(f"Compiling {len(compilation_units)} compilation units")
        files = set()
        for cu in compilation_units:
            files |= cu.files

        ctx_manager = (
            console.status(
                f"[bold green]Compiling {len(files)} files using {len(compilation_units)} solc runs...[/]"
            )
            if console
            else nullcontext()
        )
        start = time.perf_counter()

        with ctx_manager:
            # wait for compilation of all compilation units
            try:
                ret: Tuple[SolcOutput, ...] = await asyncio.gather(*tasks)
            except Exception:
                for task in tasks:
                    task.cancel()
                raise

        end = time.perf_counter()
        if console is not None:
            console.log(
                f"[green]Compiled {len(files)} files using {len(compilation_units)} solc runs in [bold green]{end - start:.2f} s[/bold green][/]"
            )

        # remove deleted files from the previous build
        for deleted_file in deleted_files:
            if deleted_file in build.source_units:
                build.reference_resolver.run_destroy_callbacks(deleted_file)
                build.source_units.pop(deleted_file)

        ctx_manager = (
            console.status(f"[bold green]Processing compilation results...[/]")
            if console
            else nullcontext()
        )
        start = time.perf_counter()

        with ctx_manager:
            successful_compilation_units = []
            all_errored_files: Set[Path] = set()

            for cu, solc_output in zip(compilation_units, ret):
                errored_files: Set[Path] = set()
                errors_per_cu[cu.hash] = set()

                for error in solc_output.errors:
                    errors_per_cu[cu.hash].add(error)
                    if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                        if error.source_location is not None:
                            path = cu.source_unit_name_to_path(
                                error.source_location.file
                            )
                            errored_files.add(path)
                        else:
                            # whole compilation unit errored
                            errored_files |= cu.files

                self._out_edge_bfs(cu, errored_files, errored_files)
                all_errored_files |= errored_files

                for file in errored_files:
                    if file in build.source_units:
                        build.reference_resolver.run_destroy_callbacks(file)
                        build.source_units.pop(file)
                    if file in build.interval_trees:
                        build.interval_trees.pop(file)

                if len(errored_files) == 0:
                    successful_compilation_units.append((cu, solc_output))

            # files_to_compile and files importing them
            files_to_recompile = set(
                graph.nodes[n[1]]["path"]  # pyright: ignore reportGeneralTypeIssues
                for n in nx.edge_bfs(
                    graph,
                    [
                        source_unit_name
                        for source_unit_name in graph.nodes  # pyright: ignore reportGeneralTypeIssues
                        if graph.nodes[source_unit_name]["path"]
                        in files_to_compile  # pyright: ignore reportGeneralTypeIssues
                    ],
                )
            ) | set(files_to_compile)

            # destroy callbacks for IR nodes that will be replaced by new ones must be executed before
            # new IR nodes are created
            for file in files_to_recompile:
                if file in build.source_units:
                    build.reference_resolver.run_destroy_callbacks(file)
                    build.source_units.pop(file)
                if file in build.interval_trees:
                    build.interval_trees.pop(file)

            # clear indexed node types responsible for handling multiple structurally different ASTs for the same file
            build.reference_resolver.clear_indexed_nodes(files_to_recompile)

            source_units_info = {
                str(source_unit): SourceUnitInfo(
                    graph.nodes[source_unit]["path"], graph.nodes[source_unit]["hash"]
                )
                for source_unit in graph.nodes
                if graph.nodes[source_unit]["path"] not in deleted_files
                and graph.nodes[source_unit]["path"] not in all_errored_files
                and graph.nodes[source_unit]["path"] not in files_to_recompile
            }

            processed_files: Set[Path] = set()
            for cu, solc_output in successful_compilation_units:
                # files requested to be compiled and files that import these files (even indirectly)
                recompiled_files: Set[Path] = set()
                self._out_edge_bfs(cu, files_to_compile & cu.files, recompiled_files)

                for source_unit_name, raw_ast in solc_output.sources.items():
                    source_units_info[source_unit_name] = SourceUnitInfo(
                        graph.nodes[source_unit_name]["path"],
                        graph.nodes[source_unit_name]["hash"],
                    )

                    path = cu.source_unit_name_to_path(source_unit_name)
                    ast = AstSolc.parse_obj(raw_ast.ast)

                    build.reference_resolver.register_source_file_id(
                        raw_ast.id, path, cu.hash
                    )
                    build.reference_resolver.index_nodes(ast, path, cu.hash)

                    if (
                        path in build.source_units and path not in recompiled_files
                    ) or path in processed_files:
                        continue
                    processed_files.add(path)
                    assert (
                        source_unit_name in graph.nodes
                    ), f"Source unit {source_unit_name} not in graph"

                    interval_tree = IntervalTree()
                    init = IrInitTuple(
                        path,
                        graph.nodes[source_unit_name]["content"].encode("utf-8"),
                        cu,
                        interval_tree,
                        build.reference_resolver,
                        solc_output.contracts[source_unit_name]
                        if source_unit_name in solc_output.contracts
                        else None,
                    )
                    build.source_units[path] = SourceUnit(init, ast)
                    build.interval_trees[path] = interval_tree

                build.reference_resolver.run_post_process_callbacks(
                    CallbackParams(
                        interval_trees=build.interval_trees,
                        source_units=build.source_units,
                    )
                )

        if console is not None:
            end = time.perf_counter()
            console.log(
                f"[green]Processed compilation results in [bold green]{end - start:.2f} s[/bold green][/]"
            )

        self._latest_build_info = ProjectBuildInfo(
            compilation_units={
                cu_hash.hex(): CompilationUnitBuildInfo(errors=list(errors))
                for cu_hash, errors in errors_per_cu.items()
            },
            allow_paths=self.__config.compiler.solc.allow_paths,
            ignore_paths=self.__config.compiler.solc.ignore_paths,
            include_paths=self.__config.compiler.solc.include_paths,
            settings=build_settings,
            source_units_info=source_units_info,
            target_solidity_version=self.__config.compiler.solc.target_version,
            woke_version=get_package_version("woke"),
            incremental=incremental,
        )
        self._latest_build = build

        if write_artifacts and (
            len(compilation_units) > 0 or len(deleted_files) > 0 or force_recompile
        ):
            logger.debug("Writing artifacts")
            self.write_artifacts(console=console)

        errors = set()
        for error_list in errors_per_cu.values():
            errors |= error_list

        logger.debug(f"Compilation finished with {len(errors)} errors")

        if console is not None:
            for error in errors:
                if (
                    error.severity == SolcOutputErrorSeverityEnum.ERROR
                    or not no_warnings
                ):
                    if error.formatted_message is not None:
                        console.print(
                            rich.panel.Panel(error.formatted_message, highlight=True)
                        )
                    else:
                        console.print(rich.panel.Panel(error.message, highlight=True))

        return build, errors

    def write_artifacts(
        self, *, console: Optional[rich.console.Console] = None
    ) -> None:
        if self._latest_build_info is None or self._latest_build is None:
            raise Exception("Project not compiled yet")

        ctx_manager = (
            console.status(f"[bold green]Writing build artifacts...[/]")
            if console
            else nullcontext()
        )
        start = time.perf_counter()

        build_key_path = self.__config.global_data_path / "build.key"
        build_key = b""
        if build_key_path.is_file():
            build_key = build_key_path.read_bytes()
        if len(build_key) != 64:
            build_key = os.urandom(64)
            build_key_path.write_bytes(build_key)

        with ctx_manager:
            build_path = self.__config.project_root_path / ".woke-build"
            build_path.mkdir(exist_ok=True)

            with (build_path / "build.json").open("w") as f:
                f.write(self._latest_build_info.json(by_alias=True, exclude_none=True))

            with (build_path / "build.bin").open("wb") as data_file, (
                build_path / "build.bin.sig"
            ).open("wb") as sig_file:
                build_data = pickle.dumps(self._latest_build)
                h = BLAKE2b.new(digest_bits=512, key=build_key)
                h.update(build_data)

                data_file.write(build_data)
                sig_file.write(h.digest())

        if console is not None:
            end = time.perf_counter()
            console.log(
                f"[green]Wrote build artifacts in [bold green]{end - start:.2f} s[/bold green][/]"
            )

    async def compile_unit_raw(
        self,
        compilation_unit: CompilationUnit,
        target_version: SolidityVersion,
        build_settings: SolcInputSettings,
    ) -> SolcOutput:
        # Dict[source_unit_name: str, path: Path]
        files = {}
        # Dict[source_unit_name: str, content: str]
        sources = {}
        for source_unit_name, data in compilation_unit.graph.nodes.items():
            path = data["path"]
            content = data["content"]
            if content is None:
                files[source_unit_name] = path
            else:
                sources[source_unit_name] = content

        # run the solc executable
        return await self.__solc_frontend.compile(
            files, sources, target_version, build_settings
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional

from intervaltree import IntervalTree

if TYPE_CHECKING:
    from woke.compiler import SolcOutputContractInfo
    from woke.compiler.compilation_unit import CompilationUnit
    from woke.ir.meta.source_unit import SourceUnit
    from woke.ir.reference_resolver import ReferenceResolver


@dataclass
class IrInitTuple:
    file: Path
    source: bytes
    cu: CompilationUnit
    interval_tree: IntervalTree
    reference_resolver: ReferenceResolver
    contracts_info: Optional[Dict[str, SolcOutputContractInfo]]
    source_unit: Optional[SourceUnit] = None



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ir.ast import (
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulLiteral,
    SolcYulSwitch,
)
from woke.ir.utils import IrInitTuple
from woke.ir.yul.abc import YulAbc, YulStatementAbc
from woke.ir.yul.case_statement import YulCase
from woke.ir.yul.function_call import YulFunctionCall
from woke.ir.yul.identifier import YulIdentifier
from woke.ir.yul.literal import YulLiteral

if TYPE_CHECKING:
    from .block import YulBlock


class YulSwitch(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _cases: List[YulCase]
    _expression: Union[YulFunctionCall, YulIdentifier, YulLiteral]

    def __init__(self, init: IrInitTuple, switch: SolcYulSwitch, parent: YulAbc):
        super().__init__(init, switch, parent)
        if isinstance(switch.expression, SolcYulFunctionCall):
            self._expression = YulFunctionCall(init, switch.expression, self)
        elif isinstance(switch.expression, SolcYulIdentifier):
            self._expression = YulIdentifier(init, switch.expression, self)
        elif isinstance(switch.expression, SolcYulLiteral):
            self._expression = YulLiteral(init, switch.expression, self)
        else:
            assert False, f"Unexpected type: {type(switch.expression)}"
        self._cases = [YulCase(init, case, self) for case in switch.cases]

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._expression
        for case_ in self._cases:
            yield from case_

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def cases(self) -> Tuple[YulCase, ...]:
        return tuple(self._cases)

    @property
    def expression(self) -> Union[YulFunctionCall, YulIdentifier, YulLiteral]:
        return self._expression



import logging
from collections import deque
from functools import lru_cache
from typing import Optional, Set, Tuple, Union

import woke.ir.types as types
from woke.ir import (
    Assignment,
    BinaryOperation,
    ExpressionAbc,
    ExpressionStatement,
    ExternalReference,
    FunctionCall,
    FunctionDefinition,
    IdentifierPathPart,
    IndexAccess,
    ModifierDefinition,
    PlaceholderStatement,
    Return,
    StatementAbc,
    TupleExpression,
    VariableDeclaration,
)
from woke.ir.enums import (
    BinaryOpOperator,
    FunctionKind,
    GlobalSymbolsEnum,
    Mutability,
    Visibility,
)
from woke.utils import recursion_guard, return_on_recursion

from .cfg import CfgBlock, ControlFlowGraph, TransitionCondition
from .utils import (
    expression_is_global_symbol,
    get_variable_declarations_from_expression,
    pair_function_call_arguments,
)

logger = logging.getLogger(__name__)


def expression_is_only_owner(
    expression: ExpressionAbc,
    msg_sender_variables: Set[VariableDeclaration],
    check_only_eoa: bool = False,
    inverted: bool = False,
) -> bool:
    if isinstance(expression, BinaryOperation):
        if expression.operator in {BinaryOpOperator.EQ, BinaryOpOperator.NEQ}:
            left = get_variable_declarations_from_expression(expression.left_expression)
            right = get_variable_declarations_from_expression(
                expression.right_expression
            )

            left_is_msg_sender = (
                len(left) == 1 and left[0] in msg_sender_variables
            ) or expression_is_global_symbol(
                expression.left_expression, GlobalSymbolsEnum.MSG_SENDER
            )
            right_is_msg_sender = (
                len(right) == 1 and right[0] in msg_sender_variables
            ) or expression_is_global_symbol(
                expression.right_expression, GlobalSymbolsEnum.MSG_SENDER
            )
            left_is_owner = not any(var is None for var in left) and all(variable_is_owner(var) for var in left)  # type: ignore
            right_is_owner = not any(var is None for var in right) and all(variable_is_owner(var) for var in right)  # type: ignore

            if check_only_eoa:
                if left_is_msg_sender and expression_is_global_symbol(
                    expression.right_expression, GlobalSymbolsEnum.TX_ORIGIN
                ):
                    return True
                if right_is_msg_sender and expression_is_global_symbol(
                    expression.left_expression, GlobalSymbolsEnum.TX_ORIGIN
                ):
                    return True

            if (left_is_msg_sender and right_is_owner) or (
                right_is_msg_sender and left_is_owner
            ):
                if (not inverted and expression.operator == BinaryOpOperator.EQ) or (
                    inverted and expression.operator == BinaryOpOperator.NEQ
                ):
                    return True
        elif expression.operator == BinaryOpOperator.BOOLEAN_OR:
            if not inverted:
                return expression_is_only_owner(
                    expression.left_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                ) and expression_is_only_owner(
                    expression.right_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                )
            else:
                return expression_is_only_owner(
                    expression.left_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                    inverted=False,
                ) or expression_is_only_owner(
                    expression.right_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                    inverted=False,
                )
        elif expression.operator == BinaryOpOperator.BOOLEAN_AND:
            if not inverted:
                return expression_is_only_owner(
                    expression.left_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                ) or expression_is_only_owner(
                    expression.right_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                )
            else:
                return expression_is_only_owner(
                    expression.left_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                    inverted=False,
                ) and expression_is_only_owner(
                    expression.right_expression,
                    msg_sender_variables,
                    check_only_eoa=check_only_eoa,
                    inverted=False,
                )
    elif isinstance(expression, FunctionCall) and isinstance(
        expression.type, types.Bool
    ):
        func = expression.function_called
        if isinstance(func, FunctionDefinition):
            if func.body is None:
                return False

            sender_variables = set(msg_sender_variables)

            logger.debug(f"{expression.source} is a function call to {func.source}")
            for arg_decl, arg_expr in pair_function_call_arguments(func, expression):
                if expression_is_global_symbol(arg_expr, GlobalSymbolsEnum.MSG_SENDER):
                    sender_variables.add(arg_decl)

            returns = []
            for statement in func.body.statements_iter():
                if isinstance(statement, Return):
                    returns.append(statement)
            if len(returns) > 0 and all(
                expression_is_only_owner(
                    return_.expression,
                    sender_variables,
                    check_only_eoa=check_only_eoa,
                    inverted=inverted,
                )
                for return_ in returns
                if return_.expression is not None
            ):
                logger.debug(f"Expression is only owner: {expression.source}")
                return True
    elif (
        isinstance(expression, IndexAccess)
        and expression.index_expression is not None
        and not inverted
    ):
        possible_owners = get_variable_declarations_from_expression(
            expression.base_expression
        )
        index_vars = get_variable_declarations_from_expression(
            expression.index_expression
        )
        index_is_msg_sender = (
            len(index_vars) == 1 and index_vars[0] in msg_sender_variables
        ) or expression_is_global_symbol(
            expression.index_expression, GlobalSymbolsEnum.MSG_SENDER
        )
        if (
            not any(var is None for var in possible_owners)
            and all(variable_is_owner(var) for var in possible_owners)  # type: ignore
            and index_is_msg_sender
        ):
            logger.debug(f"Expression is only owner: {expression.source}")
            return True
    elif (
        isinstance(expression, TupleExpression)
        and len(expression.components) == 1
        and expression.components[0] is not None
    ):
        return expression_is_only_owner(
            expression.components[0],
            msg_sender_variables,
            check_only_eoa=check_only_eoa,
            inverted=inverted,
        )

    logger.debug(f"Expression is NOT only owner: {expression.source}")
    return False


_publicly_callable_recursion_guard = set()


def statement_is_only_owner(statement: StatementAbc, check_only_eoa: bool) -> bool:
    if not isinstance(statement, ExpressionStatement):
        return False
    expr = statement.expression
    if isinstance(expr, FunctionCall):
        function_called = expr.function_called
        if function_called == GlobalSymbolsEnum.REQUIRE:
            return expression_is_only_owner(
                expr.arguments[0], set(), check_only_eoa=check_only_eoa
            )
        elif (
            isinstance(function_called, FunctionDefinition)
            and function_called.body is not None
        ):
            cfg = function_called.cfg
            assert cfg is not None
            return not _cfg_block_or_statement_is_publicly_reachable(
                cfg.end_block,
                cfg,
                check_func_visibility=False,
                check_only_eoa=check_only_eoa,
            )
    return False


_block_statement_reachable_guard = set()


@lru_cache(maxsize=2048)
def _cfg_block_or_statement_is_publicly_reachable(
    target: Union[CfgBlock, StatementAbc],
    cfg: ControlFlowGraph,
    check_func_visibility: bool = True,
    check_only_eoa: bool = False,
) -> bool:
    if (
        target,
        check_func_visibility,
        check_only_eoa,
    ) in _block_statement_reachable_guard:
        return False

    with recursion_guard(
        _block_statement_reachable_guard, target, check_func_visibility, check_only_eoa
    ):
        decl = cfg.declaration

        if isinstance(decl, FunctionDefinition):
            if decl.kind == FunctionKind.CONSTRUCTOR:
                return False
            for modifier in decl.modifiers:
                mod = modifier.modifier_name.referenced_declaration
                assert isinstance(mod, ModifierDefinition)
                if mod.body is None:
                    continue
                placeholders = [
                    s
                    for s in mod.body.statements_iter()
                    if isinstance(s, PlaceholderStatement)
                ]
                if not any(
                    statement_is_publicly_executable(s, check_only_eoa=check_only_eoa)
                    for s in placeholders
                ):
                    return False

        if isinstance(target, StatementAbc):
            block = cfg.get_cfg_block(target)
            if target not in block.statements:
                assert target == block.control_statement
                if any(
                    statement_is_only_owner(s, check_only_eoa) for s in block.statements
                ):
                    return False
            else:
                index = block.statements.index(target)
                if any(
                    statement_is_only_owner(s, check_only_eoa)
                    for s in block.statements[:index]
                ):
                    return False
        elif isinstance(target, CfgBlock):
            block = target
        else:
            raise NotImplementedError()

        if block == cfg.start_block:
            reached_start = True
        else:
            graph = cfg.graph
            visited = {block}
            queue = deque([block])
            reached_start = False

            while len(queue) > 0:
                block = queue.popleft()
                from_: CfgBlock
                condition: Tuple[TransitionCondition, Optional[ExpressionAbc]]
                for (
                    from_,
                    _,
                    data,
                ) in graph.in_edges(  # pyright: ignore reportGeneralTypeIssues
                    block, data=True  # pyright: ignore reportGeneralTypeIssues
                ):
                    if from_ in visited:
                        continue
                    condition = data[  # pyright: ignore reportOptionalSubscript
                        "condition"
                    ]
                    if (
                        (condition[0] == TransitionCondition.NEVER)
                        or (
                            condition[0] == TransitionCondition.IS_TRUE
                            and condition[1] is not None
                            and expression_is_only_owner(
                                condition[1],  # pyright: ignore reportGeneralTypeIssues
                                set(),
                                check_only_eoa=check_only_eoa,
                            )
                        )
                        or (
                            condition[0] == TransitionCondition.IS_FALSE
                            and condition[1] is not None
                            and expression_is_only_owner(
                                condition[1],  # pyright: ignore reportGeneralTypeIssues
                                set(),
                                check_only_eoa=check_only_eoa,
                                inverted=True,
                            )
                        )
                    ):
                        continue

                    if any(
                        statement_is_only_owner(s, check_only_eoa)
                        for s in from_.statements
                    ):
                        continue

                    if from_ == cfg.start_block:
                        reached_start = True
                        break

                    visited.add(from_)
                    queue.append(from_)

        if not reached_start:
            return False
        if isinstance(decl, ModifierDefinition) or not check_func_visibility:
            return True
        if decl.visibility in {Visibility.PUBLIC, Visibility.EXTERNAL}:
            return True

        for ref in decl.get_all_references(False):
            if isinstance(ref, IdentifierPathPart):
                calling_statement = ref.underlying_node
            elif isinstance(ref, ExternalReference):
                continue  # TODO currently not supported
            else:
                calling_statement = ref
            while calling_statement is not None:
                if isinstance(calling_statement, StatementAbc):
                    break
                calling_statement = calling_statement.parent
            if calling_statement is not None and statement_is_publicly_executable(
                calling_statement, check_only_eoa=check_only_eoa
            ):
                return True
        return False


@lru_cache(maxsize=2048)
def statement_is_publicly_executable(
    statement: StatementAbc,
    check_func_visibility: bool = True,
    check_only_eoa: bool = False,
) -> bool:
    decl = statement
    while decl is not None:
        if isinstance(decl, (FunctionDefinition, ModifierDefinition)):
            break
        decl = decl.parent
    assert isinstance(decl, (FunctionDefinition, ModifierDefinition))

    cfg = decl.cfg
    assert cfg is not None
    return _cfg_block_or_statement_is_publicly_reachable(
        statement,
        cfg,
        check_func_visibility=check_func_visibility,
        check_only_eoa=check_only_eoa,
    )


@return_on_recursion(True)
def variable_is_owner(variable: VariableDeclaration) -> bool:
    is_mapping_owners = (
        isinstance(variable.type, types.Mapping)
        and isinstance(variable.type.key_type, types.Address)
        and isinstance(variable.type.value_type, types.Bool)
    )

    if (
        not isinstance(variable.type, (types.Address, types.Contract))
        and not is_mapping_owners
    ):
        return False
    if variable.mutability in {Mutability.CONSTANT, Mutability.IMMUTABLE}:
        return True
    if not variable.is_state_variable:
        return False

    for ref in variable.references:
        # ref is assignment?
        if isinstance(ref, IdentifierPathPart):
            node = ref.underlying_node
        elif isinstance(ref, ExternalReference):
            continue  # TODO currently not supported
        else:
            node = ref
        is_assignment = False
        while node is not None:
            if isinstance(node, Assignment):
                for assigned_paths in node.assigned_variables:
                    if assigned_paths is not None:
                        for assigned_path in assigned_paths:
                            if variable in assigned_path:
                                is_assignment = True
                break
            node = node.parent

        if not is_assignment:
            continue

        if isinstance(ref, IdentifierPathPart):
            node = ref.underlying_node
        elif isinstance(ref, ExternalReference):
            continue  # TODO currently not supported
        else:
            node = ref
        while node is not None:
            if isinstance(node, StatementAbc):
                break
            node = node.parent
        if node is not None and statement_is_publicly_executable(node):
            logger.debug(f"Variable is NOT owner: {variable.source} {node.source}")
            return False

    return True



# tab space width for indentation
TAB_WIDTH = 4

DEFAULT_IMPORTS: str = """
from __future__ import annotations

import dataclasses
from typing import List, Dict, Optional, overload, Union, Callable, Tuple
from typing_extensions import Literal

from woke.development.core import Contract, Library, Address, Account, Chain, RequestType
from woke.development.primitive_types import *
from woke.development.transactions import TransactionAbc, TransactionRevertedError
"""

INIT_CONTENT: str = """
import woke.development.core
from woke.utils import get_package_version

if get_package_version("woke") != "{version}":
    raise RuntimeError("Pytypes generated for a different version of woke. Please regenerate.")

woke.development.core.errors = {errors}
woke.development.core.events = {events}
woke.development.core.contracts_by_fqn = {contracts_by_fqn}
woke.development.core.contracts_by_metadata = {contracts_by_metadata}
woke.development.core.contracts_inheritance = {contracts_inheritance}
woke.development.core.contracts_revert_index = {contracts_revert_index}
woke.development.core.creation_code_index = {creation_code_index}
"""



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Optional, Union

from ..expressions.abc import ExpressionAbc
from ..type_names.user_defined_type_name import UserDefinedTypeName
from ..utils import IrInitTuple
from .identifier_path import IdentifierPath

if TYPE_CHECKING:
    from ..declarations.contract_definition import ContractDefinition

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import (
    SolcIdentifierPath,
    SolcInheritanceSpecifier,
    SolcUserDefinedTypeName,
)


class InheritanceSpecifier(SolidityAbc):
    """
    !!! example
        `:::solidity A(0x1234567890123456789012345678901234567890)` in the following code:
        ```solidity
        contract A {
            address immutable owner;

            constructor(address _owner) {
                owner = _owner;
            }
        }

        contract B is A(0x1234567890123456789012345678901234567890) {}
        ```
    """

    _ast_node: SolcInheritanceSpecifier
    _parent: ContractDefinition

    _base_name: Union[IdentifierPath, UserDefinedTypeName]
    _arguments: Optional[List[ExpressionAbc]]

    def __init__(
        self,
        init: IrInitTuple,
        inheritance_specifier: SolcInheritanceSpecifier,
        parent: ContractDefinition,
    ):
        super().__init__(init, inheritance_specifier, parent)

        if isinstance(inheritance_specifier.base_name, SolcIdentifierPath):
            self._base_name = IdentifierPath(
                init, inheritance_specifier.base_name, self
            )
        elif isinstance(inheritance_specifier.base_name, SolcUserDefinedTypeName):
            self._base_name = UserDefinedTypeName(
                init, inheritance_specifier.base_name, self
            )

        if inheritance_specifier.arguments is None:
            self._arguments = None
        else:
            self._arguments = []
            for argument in inheritance_specifier.arguments:
                self._arguments.append(ExpressionAbc.from_ast(init, argument, self))

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._base_name
        if self._arguments is not None:
            for argument in self._arguments:
                yield from argument

    @property
    def parent(self) -> ContractDefinition:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def base_name(self) -> Union[IdentifierPath, UserDefinedTypeName]:
        """
        The returned IR node holds a reference to the base contract definition ([ContractDefinition][woke.ir.declarations.contract_definition.ContractDefinition]).
        !!! example
            `A` in the following code:
            ```solidity
            contract B is A(0x1234567890123456789012345678901234567890) {}
            ```
        Returns:
            IR node representing the base contract name.
        """
        return self._base_name

    @property
    def arguments(self) -> Optional[List[ExpressionAbc]]:
        """
        !!! warning
            Is `None` when there are no curly braces after the inheritance specifier name.
            ```solidity
            contract B is A {}
            ```

            Is an empty list when there are curly braces but no arguments.
            ```solidity
            contract B is A() {}
            ```
        !!! example
            `:::solidity 0x1234567890123456789012345678901234567890` in the following code:
            ```solidity
            contract B is A(0x1234567890123456789012345678901234567890) {}
            ```
        Returns:
            Arguments of the base constructor call, if provided.
        """
        return self._arguments



from rich import print

from woke.development.core import Abi, Account, Address, Eip712Domain, Wei
from woke.development.internal import UnknownEvent
from woke.development.primitive_types import *
from woke.development.transactions import (
    Eip1559Transaction,
    Eip2930Transaction,
    Error,
    LegacyTransaction,
    Panic,
    PanicCodeEnum,
    TransactionAbc,
    TransactionRevertedError,
    UnknownTransactionRevertedError,
    may_revert,
    must_revert,
    on_revert,
)
from woke.development.utils import (
    get_create2_address_from_code,
    get_create2_address_from_hash,
    get_create_address,
    keccak256,
)

from .core import Chain, default_chain



TEMPLATE = """from __future__ import annotations

from typing import List

import rich_click as click
import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class {class_name}(Detector):
    def detect(self) -> List[DetectorResult]:
        return []

    @detector.command(name="{command_name}")
    def cli(self) -> None:
        pass
"""



from functools import lru_cache
from typing import Iterator, Optional, Set, Tuple, Union

from typing_extensions import Literal

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcAssignment
from woke.ir.enums import AssignmentOperator, GlobalSymbolsEnum, ModifiesStateFlag
from woke.ir.utils import IrInitTuple

from ..declarations.abc import DeclarationAbc
from ..declarations.function_definition import FunctionDefinition
from ..declarations.struct_definition import StructDefinition
from ..declarations.variable_declaration import VariableDeclaration
from ..meta.source_unit import SourceUnit
from .abc import ExpressionAbc
from .conditional import Conditional
from .function_call import FunctionCall
from .identifier import Identifier
from .index_access import IndexAccess
from .member_access import MemberAccess
from .tuple_expression import TupleExpression

AssignedVariablePath = Tuple[
    Union[DeclarationAbc, SourceUnit, Literal["IndexAccess"]], ...
]


class Assignment(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcAssignment
    _parent: SolidityAbc  # TODO: make this more specific

    _left_expression: ExpressionAbc
    _right_expression: ExpressionAbc
    _operator: AssignmentOperator

    def __init__(
        self, init: IrInitTuple, assignment: SolcAssignment, parent: SolidityAbc
    ):
        super().__init__(init, assignment, parent)
        self._operator = assignment.operator
        self._left_expression = ExpressionAbc.from_ast(
            init, assignment.left_hand_side, self
        )
        self._right_expression = ExpressionAbc.from_ast(
            init, assignment.right_hand_side, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._left_expression
        yield from self._right_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def left_expression(self) -> ExpressionAbc:
        return self._left_expression

    @property
    def right_expression(self) -> ExpressionAbc:
        return self._right_expression

    @property
    def operator(self) -> AssignmentOperator:
        return self._operator

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        return self.left_expression.is_ref_to_state_variable

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.left_expression.modifies_state | self.right_expression.modifies_state
        if self.left_expression.is_ref_to_state_variable:
            ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
        return ret

    @property
    @lru_cache(maxsize=2048)
    def assigned_variables(self) -> Tuple[Optional[Set[AssignedVariablePath]], ...]:
        def resolve_node(node: ExpressionAbc) -> Set[AssignedVariablePath]:
            if isinstance(node, Conditional):
                return resolve_node(node.true_expression) | resolve_node(
                    node.false_expression
                )
            elif isinstance(node, Identifier):
                referenced_declaration = node.referenced_declaration
                assert isinstance(referenced_declaration, (DeclarationAbc, SourceUnit))
                return {(referenced_declaration,)}
            elif isinstance(node, IndexAccess):
                return {
                    path + ("IndexAccess",)
                    for path in resolve_node(node.base_expression)
                }
            elif isinstance(node, MemberAccess):
                referenced_declaration = node.referenced_declaration
                assert isinstance(referenced_declaration, (DeclarationAbc, SourceUnit))
                return {
                    path + (referenced_declaration,)
                    for path in resolve_node(node.expression)
                }
            elif isinstance(node, FunctionCall):
                function_called = node.function_called
                if function_called is None:
                    return set()
                elif isinstance(
                    function_called, (GlobalSymbolsEnum, VariableDeclaration)
                ):
                    # global function or variable getter called
                    # variable getter may return different type than variable declaration (structs with arrays and mappings)
                    # return empty set for now
                    return set()
                elif isinstance(function_called, FunctionDefinition):
                    # cannot be handled in the current implementation, return empty set for now
                    return set()
                elif isinstance(function_called, StructDefinition):
                    return {(function_called,)}
                else:
                    assert False, f"Unexpected node type: {type(node)}\n{self.source}"
            else:
                assert False, f"Unexpected node type: {type(node)}\n{self.source}"

        node = self.left_expression
        if isinstance(node, TupleExpression):
            return tuple(
                resolve_node(expression) if expression is not None else None
                for expression in node.components
            )
        else:
            return (resolve_node(node),)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ir.ast import (
    SolcYulAssignment,
    SolcYulBlock,
    SolcYulBreak,
    SolcYulContinue,
    SolcYulExpressionStatement,
    SolcYulForLoop,
    SolcYulFunctionDefinition,
    SolcYulIf,
    SolcYulLeave,
    SolcYulSwitch,
    SolcYulVariableDeclaration,
)
from woke.ir.utils import IrInitTuple

from .abc import YulAbc, YulStatementAbc

if TYPE_CHECKING:
    from woke.ir.statements.inline_assembly import InlineAssembly

    from .assignment import YulAssignment
    from .break_statement import YulBreak
    from .case_statement import YulCase
    from .continue_statement import YulContinue
    from .expression_statement import YulExpressionStatement
    from .for_loop import YulForLoop
    from .function_definition import YulFunctionDefinition
    from .if_statement import YulIf
    from .leave import YulLeave
    from .switch import YulSwitch
    from .variable_declaration import YulVariableDeclaration


class YulBlock(YulStatementAbc):
    """
    TBD
    """

    _parent: Union[
        InlineAssembly, YulBlock, YulForLoop, YulFunctionDefinition, YulIf, YulCase
    ]
    _statements: List[
        Union[
            YulAssignment,
            "YulBlock",
            YulBreak,
            YulContinue,
            YulExpressionStatement,
            YulLeave,
            YulForLoop,
            YulFunctionDefinition,
            YulIf,
            YulSwitch,
            YulVariableDeclaration,
        ]
    ]

    def __init__(
        self,
        init: IrInitTuple,
        block: SolcYulBlock,
        parent: Union[InlineAssembly, YulAbc],
    ):
        from .assignment import YulAssignment
        from .break_statement import YulBreak
        from .continue_statement import YulContinue
        from .expression_statement import YulExpressionStatement
        from .for_loop import YulForLoop
        from .function_definition import YulFunctionDefinition
        from .if_statement import YulIf
        from .leave import YulLeave
        from .switch import YulSwitch
        from .variable_declaration import YulVariableDeclaration

        super().__init__(init, block, parent)
        self._statements = []
        for statement in block.statements:
            if isinstance(statement, SolcYulAssignment):
                self._statements.append(YulAssignment(init, statement, self))
            elif isinstance(statement, SolcYulBlock):
                self._statements.append(YulBlock(init, statement, self))
            elif isinstance(statement, SolcYulBreak):
                self._statements.append(YulBreak(init, statement, self))
            elif isinstance(statement, SolcYulContinue):
                self._statements.append(YulContinue(init, statement, self))
            elif isinstance(statement, SolcYulExpressionStatement):
                self._statements.append(YulExpressionStatement(init, statement, self))
            elif isinstance(statement, SolcYulLeave):
                self._statements.append(YulLeave(init, statement, self))
            elif isinstance(statement, SolcYulForLoop):
                self._statements.append(YulForLoop(init, statement, self))
            elif isinstance(statement, SolcYulFunctionDefinition):
                self._statements.append(YulFunctionDefinition(init, statement, self))
            elif isinstance(statement, SolcYulIf):
                self._statements.append(YulIf(init, statement, self))
            elif isinstance(statement, SolcYulSwitch):
                self._statements.append(YulSwitch(init, statement, self))
            elif isinstance(statement, SolcYulVariableDeclaration):
                self._statements.append(YulVariableDeclaration(init, statement, self))
            else:
                assert False, f"Unexpected type: {type(statement)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        for statement in self._statements:
            yield from statement

    @property
    def parent(
        self,
    ) -> Union[
        InlineAssembly, YulBlock, YulForLoop, YulFunctionDefinition, YulIf, YulCase
    ]:
        return self._parent

    @property
    def statements(
        self,
    ) -> Tuple[
        Union[
            YulAssignment,
            "YulBlock",
            YulBreak,
            YulContinue,
            YulExpressionStatement,
            YulLeave,
            YulForLoop,
            YulFunctionDefinition,
            YulIf,
            YulSwitch,
            YulVariableDeclaration,
        ],
        ...,
    ]:
        return tuple(self._statements)



from woke.cli.detect import run_detect as detector

from .api import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
)
from .balance_state_var import UnsafeAddressBalanceUseDetector
from .bug_empty_byte_array_copy import BugEmptyByteArrayCopyDetector
from .call_options_not_called import FunctionCallOptionsNotCalledDetector
from .missing_return import MissingReturnDetector
from .msg_value_nonpayable_function import MsgValueNonpayableFunctionDetector
from .overflow_calldata_tuple_reencoding_bug import (
    OverflowCalldataTupleReencodingBugDetector,
)
from .proxy_contract_selector_clashes import ProxyContractSelectorClashDetector
from .reentrancy import ReentrancyDetector
from .unchecked_return_value import UncheckedReturnValueDetector
from .unsafe_delegatecall import UnsafeDelegatecallDetector
from .unsafe_selfdestruct import UnsafeSelfdestructDetector
from .unsafe_tx_origin import UnsafeTxOriginDetector
from .unused_contract import UnusedContractDetector

from . import axelar  # isort:skip



from __future__ import annotations

import heapq
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
)

from intervaltree import IntervalTree

from woke.ir.enums import GlobalSymbolsEnum

if TYPE_CHECKING:
    from woke.ir.abc import SolidityAbc
    from woke.ir.expressions.identifier import Identifier
    from woke.ir.expressions.member_access import MemberAccess
    from woke.ir.meta.source_unit import SourceUnit

from woke.ir.ast import AstNodeId, AstSolc

logger = logging.getLogger(__name__)


@dataclass
class CallbackParams:
    interval_trees: Dict[Path, IntervalTree]
    source_units: Dict[Path, SourceUnit]


@dataclass(order=True)
class PostProcessQueueItem:
    priority: int
    callback: Callable[[CallbackParams], None] = field(compare=False)


class ReferenceResolver:
    _ordered_nodes: DefaultDict[bytes, Dict[AstNodeId, Tuple[Path, int]]]
    _ordered_nodes_inverted: DefaultDict[bytes, Dict[Tuple[Path, int], AstNodeId]]
    _registered_source_files: DefaultDict[bytes, Dict[int, Path]]
    _registered_nodes: Dict[Tuple[Path, int], SolidityAbc]
    _post_process_callbacks: List[PostProcessQueueItem]
    _destroy_callbacks: DefaultDict[Path, List[Callable[[], None]]]
    _global_symbol_references: DefaultDict[
        GlobalSymbolsEnum, List[Union[Identifier, MemberAccess]]
    ]
    _node_types: Dict[Path, Dict[int, str]]

    def __init__(self):
        self._ordered_nodes = defaultdict(dict)
        self._ordered_nodes_inverted = defaultdict(dict)
        self._registered_source_files = defaultdict(dict)
        self._registered_nodes = {}
        self._post_process_callbacks = []
        self._destroy_callbacks = defaultdict(list)
        self._global_symbol_references = defaultdict(list)
        self._node_types = {}

    def clear_indexed_nodes(self, paths: Iterable[Path]) -> None:
        for path in paths:
            self._node_types.pop(path, None)

    def index_nodes(self, root_node: AstSolc, path: Path, cu_hash: bytes) -> None:
        if path not in self._node_types:
            self._node_types[path] = {}
            self._node_types[path][0] = root_node.node_type
            check = False
        else:
            assert self._node_types[path][0] == root_node.node_type
            check = True
        prev_type = root_node.node_type

        self._ordered_nodes[cu_hash][root_node.id] = (path, 0)
        self._ordered_nodes_inverted[cu_hash][(path, 0)] = root_node.id
        index = 1
        for node in root_node:
            if check:
                skip = False
                other_type = self._node_types[path].get(index)

                while other_type != node.node_type:
                    if other_type == "StructuredDocumentation":
                        index += 1
                        other_type = self._node_types[path].get(index)
                        continue
                    elif node.node_type == "StructuredDocumentation":
                        skip = True
                        prev_type = "StructuredDocumentation"
                        break
                    elif (
                        self._node_types[path][index - 1] == "UserDefinedTypeName"
                        and prev_type == "UserDefinedTypeName"
                    ):
                        if other_type == "IdentifierPath":
                            index += 1
                            other_type = self._node_types[path].get(index)
                            continue
                        elif node.node_type == "IdentifierPath":
                            skip = True
                            prev_type = "IdentifierPath"
                            break
                    elif (
                        other_type == "IdentifierPath"
                        and node.node_type == "UserDefinedTypeName"
                    ) or (
                        other_type == "UserDefinedTypeName"
                        and node.node_type == "IdentifierPath"
                    ):
                        break

                    assert (
                        other_type == node.node_type
                    ), f"Expected {other_type} but got {node.node_type} at {path}:{index} {node.id}"

                if skip:
                    continue
            else:
                self._node_types[path][index] = node.node_type

            self._ordered_nodes[cu_hash][node.id] = (path, index)
            self._ordered_nodes_inverted[cu_hash][(path, index)] = node.id

            prev_type = node.node_type
            index += 1

    def register_source_file_id(self, source_file_id: int, path: Path, cu_hash: bytes):
        self._registered_source_files[cu_hash][source_file_id] = path

    def get_node_path_order(
        self, node_id: AstNodeId, cu_hash: bytes
    ) -> Tuple[Path, int]:
        return self._ordered_nodes[cu_hash][node_id]

    def get_ast_id_from_cu_node_path_order(
        self, node_path_order: Tuple[Path, int], cu_hash: bytes
    ) -> AstNodeId:
        return self._ordered_nodes_inverted[cu_hash][node_path_order]

    def register_node(self, node: SolidityAbc, node_id: AstNodeId, cu_hash: bytes):
        assert cu_hash in self._ordered_nodes
        assert node_id in self._ordered_nodes[cu_hash]
        node_path_order = self._ordered_nodes[cu_hash][node_id]
        self._registered_nodes[node_path_order] = node

    def resolve_node(self, node_id: AstNodeId, cu_hash: bytes) -> SolidityAbc:
        node_path_order = self._ordered_nodes[cu_hash][node_id]
        return self._registered_nodes[node_path_order]

    def resolve_source_file_id(self, source_file_id: int, cu_hash: bytes) -> Path:
        return self._registered_source_files[cu_hash][source_file_id]

    def register_post_process_callback(
        self, callback: Callable[[CallbackParams], None], priority: int = 0
    ):
        heapq.heappush(
            self._post_process_callbacks, PostProcessQueueItem(priority, callback)
        )

    def register_destroy_callback(self, file: Path, callback: Callable[[], None]):
        self._destroy_callbacks[file].append(callback)

    def run_post_process_callbacks(self, callback_params: CallbackParams):
        while len(self._post_process_callbacks):
            callback = heapq.heappop(self._post_process_callbacks).callback
            callback(callback_params)

    def run_destroy_callbacks(self, file: Path):
        for callback in self._destroy_callbacks[file]:
            callback()
        del self._destroy_callbacks[file]

    def register_global_symbol_reference(
        self, node_id: GlobalSymbolsEnum, node: Union[Identifier, MemberAccess]
    ):
        self._global_symbol_references[node_id].append(node)

    def unregister_global_symbol_reference(
        self, node_id: GlobalSymbolsEnum, node: Union[Identifier, MemberAccess]
    ):
        self._global_symbol_references[node_id].remove(node)

    def get_global_symbol_references(
        self, node_id: GlobalSymbolsEnum
    ) -> Tuple[Union[Identifier, MemberAccess], ...]:
        return tuple(self._global_symbol_references[node_id])
